/*-
 * Copyright (c) 2009-2013 The FreeBSD Foundation
 * All rights reserved.
 *
 * This software was developed by Pawel Jakub Dawidek under sponsorship from
 * the FreeBSD Foundation.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#ifndef	_NV_H_
#define	_NV_H_

#include <sys/cdefs.h>

#include <stdarg.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>

#ifndef	_NVPAIR_T_DECLARED
#define	_NVPAIR_T_DECLARED
struct nvpair;

typedef struct nvpair nvpair_t;
#endif

#ifndef	_NVLIST_T_DECLARED
#define	_NVLIST_T_DECLARED
struct nvlist;

typedef struct nvlist nvlist_t;
#endif

#define	NVPAIR_NAME_MAX	256

#define	NV_TYPE_NONE			-1

#define	NV_TYPE_NULL			0
#define	NV_TYPE_BOOL			1
#define	NV_TYPE_INT8			2
#define	NV_TYPE_UINT8			3
#define	NV_TYPE_INT16			4
#define	NV_TYPE_UINT16			5
#define	NV_TYPE_INT32			6
#define	NV_TYPE_UINT32			7
#define	NV_TYPE_INT64			8
#define	NV_TYPE_UINT64			9
#define	NV_TYPE_STRING			10
#define	NV_TYPE_NVLIST			11
#define	NV_TYPE_DESCRIPTOR		12
#define	NV_TYPE_BOOL_ARRAY		13
#define	NV_TYPE_INT8_ARRAY		14
#define	NV_TYPE_UINT8_ARRAY		15
#define	NV_TYPE_INT16_ARRAY		16
#define	NV_TYPE_UINT16_ARRAY		17
#define	NV_TYPE_INT32_ARRAY		18
#define	NV_TYPE_UINT32_ARRAY		19
#define	NV_TYPE_INT64_ARRAY		20
#define	NV_TYPE_UINT64_ARRAY		21
#define	NV_TYPE_STRING_ARRAY		22
#define	NV_TYPE_NVLIST_ARRAY		23
#define	NV_TYPE_DESCRIPTOR_ARRAY	24

/* Duplicated names are not allowed. */
#define	NV_FLAG_UNIQUE_NAME		0x01
/* Duplicated names of the same type are not allowed. */
#define	NV_FLAG_UNIQUE_NAME_TYPE	0x02
/*
 * NULL values for arrays and strings as well as -1 for descriptors
 * are allowed. Note that individual entries in nvlist array, string
 * array or descriptor array can still be NULL/-1.
 */
#define	NV_FLAG_NO_VALUE		0x04

nvlist_t *nvlist_create(int flags);
void nvlist_destroy(nvlist_t *nvl);
int nvlist_error(const nvlist_t *nvl);

bool nvlist_exists_type(const nvlist_t *nvl, const char *name, int type);
bool nvlist_existsf_type(const nvlist_t *nvl, int type, const char *namefmt, ...) __printflike(3, 4);
bool nvlist_existsv_type(const nvlist_t *nvl, int type, const char *namefmt, va_list nameap) __printflike(3, 0);

int nvlist_free_type(nvlist_t *nvl, const char *name, int type);
int nvlist_freef_type(nvlist_t *nvl, int type, const char *namefmt, ...) __printflike(3, 4);
int nvlist_freev_type(nvlist_t *nvl, int type, const char *namefmt, va_list nameap) __printflike(3, 0);

nvlist_t *nvlist_clone(const nvlist_t *nvl);

void nvlist_dump(const nvlist_t *nvl, int fd);
void nvlist_fdump(const nvlist_t *nvl, FILE *out);

size_t nvlist_size(const nvlist_t *nvl);
int *nvlist_descriptors(const nvlist_t *nvl, size_t *nitemsp);
size_t nvlist_ndescriptors(const nvlist_t *nvl);
void *nvlist_pack(const nvlist_t *nvl, size_t *sizep);
nvlist_t *nvlist_unpack(const void *buf, size_t size);

int nvlist_send(int sock, const nvlist_t *nvl);
nvlist_t *nvlist_recv(int sock);
nvlist_t *nvlist_xfer(int sock, nvlist_t *nvl);

nvpair_t *nvlist_first_nvpair(const nvlist_t *nvl);
nvpair_t *nvlist_next_nvpair(const nvlist_t *nvl, const nvpair_t *nvp);
nvpair_t *nvlist_prev_nvpair(const nvlist_t *nvl, const nvpair_t *nvp);

/*
 * The nvlist_exists functions check if the given name (optionally of the given
 * type) exists on nvlist.
 */

bool nvlist_exists(const nvlist_t *nvl, const char *name);
bool nvlist_exists_null(const nvlist_t *nvl, const char *name);
bool nvlist_exists_bool(const nvlist_t *nvl, const char *name);
bool nvlist_exists_int8(const nvlist_t *nvl, const char *name);
bool nvlist_exists_uint8(const nvlist_t *nvl, const char *name);
bool nvlist_exists_int16(const nvlist_t *nvl, const char *name);
bool nvlist_exists_uint16(const nvlist_t *nvl, const char *name);
bool nvlist_exists_int32(const nvlist_t *nvl, const char *name);
bool nvlist_exists_uint32(const nvlist_t *nvl, const char *name);
bool nvlist_exists_int64(const nvlist_t *nvl, const char *name);
bool nvlist_exists_uint64(const nvlist_t *nvl, const char *name);
bool nvlist_exists_string(const nvlist_t *nvl, const char *name);
bool nvlist_exists_nvlist(const nvlist_t *nvl, const char *name);
bool nvlist_exists_descriptor(const nvlist_t *nvl, const char *name);
bool nvlist_exists_bool_array(const nvlist_t *nvl, const char *name);
bool nvlist_exists_int8_array(const nvlist_t *nvl, const char *name);
bool nvlist_exists_uint8_array(const nvlist_t *nvl, const char *name);
bool nvlist_exists_int16_array(const nvlist_t *nvl, const char *name);
bool nvlist_exists_uint16_array(const nvlist_t *nvl, const char *name);
bool nvlist_exists_int32_array(const nvlist_t *nvl, const char *name);
bool nvlist_exists_uint32_array(const nvlist_t *nvl, const char *name);
bool nvlist_exists_int64_array(const nvlist_t *nvl, const char *name);
bool nvlist_exists_uint64_array(const nvlist_t *nvl, const char *name);
bool nvlist_exists_string_array(const nvlist_t *nvl, const char *name);
bool nvlist_exists_nvlist_array(const nvlist_t *nvl, const char *name);
bool nvlist_exists_descriptor_array(const nvlist_t *nvl, const char *name);

bool nvlist_existsf(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool nvlist_existsf_null(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool nvlist_existsf_bool(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool nvlist_existsf_int8(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool nvlist_existsf_uint8(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool nvlist_existsf_int16(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool nvlist_existsf_uint16(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool nvlist_existsf_int32(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool nvlist_existsf_uint32(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool nvlist_existsf_int64(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool nvlist_existsf_uint64(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool nvlist_existsf_string(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool nvlist_existsf_nvlist(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool nvlist_existsf_descriptor(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool nvlist_existsf_bool_array(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool nvlist_existsf_int8_array(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool nvlist_existsf_uint8_array(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool nvlist_existsf_int16_array(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool nvlist_existsf_uint16_array(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool nvlist_existsf_int32_array(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool nvlist_existsf_uint32_array(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool nvlist_existsf_int64_array(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool nvlist_existsf_uint64_array(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool nvlist_existsf_string_array(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool nvlist_existsf_nvlist_array(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool nvlist_existsf_descriptor_array(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);

bool nvlist_existsv(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool nvlist_existsv_null(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool nvlist_existsv_bool(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool nvlist_existsv_int8(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool nvlist_existsv_uint8(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool nvlist_existsv_int16(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool nvlist_existsv_uint16(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool nvlist_existsv_int32(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool nvlist_existsv_uint32(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool nvlist_existsv_int64(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool nvlist_existsv_uint64(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool nvlist_existsv_string(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool nvlist_existsv_nvlist(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool nvlist_existsv_descriptor(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool nvlist_existsv_bool_array(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool nvlist_existsv_int8_array(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool nvlist_existsv_uint8_array(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool nvlist_existsv_int16_array(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool nvlist_existsv_uint16_array(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool nvlist_existsv_int32_array(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool nvlist_existsv_uint32_array(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool nvlist_existsv_int64_array(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool nvlist_existsv_uint64_array(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool nvlist_existsv_string_array(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool nvlist_existsv_nvlist_array(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool nvlist_existsv_descriptor_array(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);

/*
 * The nvlist_add functions add the given name/value pair.
 * If a pointer is provided, nvlist_add will internally allocate memory for the
 * given data (in other words it won't consume provided buffer).
 */

void nvlist_add_nvpair(nvlist_t *nvl, nvpair_t *nvp);
void nvlist_add_null(nvlist_t *nvl, const char *name);
void nvlist_add_bool(nvlist_t *nvl, const char *name, bool value);
void nvlist_add_int8(nvlist_t *nvl, const char *name, int8_t value);
void nvlist_add_uint8(nvlist_t *nvl, const char *name, uint8_t value);
void nvlist_add_int16(nvlist_t *nvl, const char *name, int16_t value);
void nvlist_add_uint16(nvlist_t *nvl, const char *name, uint16_t value);
void nvlist_add_int32(nvlist_t *nvl, const char *name, int32_t value);
void nvlist_add_uint32(nvlist_t *nvl, const char *name, uint32_t value);
void nvlist_add_int64(nvlist_t *nvl, const char *name, int64_t value);
void nvlist_add_uint64(nvlist_t *nvl, const char *name, uint64_t value);
void nvlist_add_string(nvlist_t *nvl, const char *name, const char *value);
void nvlist_add_stringf(nvlist_t *nvl, const char *name, const char *valuefmt, ...) __printflike(3, 4);
void nvlist_add_stringv(nvlist_t *nvl, const char *name, const char *valuefmt, va_list valueap) __printflike(3, 0);
void nvlist_add_nvlist(nvlist_t *nvl, const char *name, const nvlist_t *value);
void nvlist_add_descriptor(nvlist_t *nvl, const char *name, int value);
void nvlist_add_bool_array(nvlist_t *nvl, const char *name, const bool *value, size_t nitems);
void nvlist_add_int8_array(nvlist_t *nvl, const char *name, const int8_t *value, size_t nitems);
void nvlist_add_uint8_array(nvlist_t *nvl, const char *name, const uint8_t *value, size_t nitems);
void nvlist_add_int16_array(nvlist_t *nvl, const char *name, const int16_t *value, size_t nitems);
void nvlist_add_uint16_array(nvlist_t *nvl, const char *name, const uint16_t *value, size_t nitems);
void nvlist_add_int32_array(nvlist_t *nvl, const char *name, const int32_t *value, size_t nitems);
void nvlist_add_uint32_array(nvlist_t *nvl, const char *name, const uint32_t *value, size_t nitems);
void nvlist_add_int64_array(nvlist_t *nvl, const char *name, const int64_t *value, size_t nitems);
void nvlist_add_uint64_array(nvlist_t *nvl, const char *name, const uint64_t *value, size_t nitems);
void nvlist_add_string_array(nvlist_t *nvl, const char *name, const char * const *value, size_t nitems);
void nvlist_add_nvlist_array(nvlist_t *nvl, const char *name, const nvlist_t * const *value, size_t nitems);
void nvlist_add_descriptor_array(nvlist_t *nvl, const char *name, const int *value, size_t nitems);

void nvlist_addf_null(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
void nvlist_addf_bool(nvlist_t *nvl, bool value, const char *namefmt, ...) __printflike(3, 4);
void nvlist_addf_int8(nvlist_t *nvl, int8_t value, const char *namefmt, ...) __printflike(3, 4);
void nvlist_addf_uint8(nvlist_t *nvl, uint8_t value, const char *namefmt, ...) __printflike(3, 4);
void nvlist_addf_int16(nvlist_t *nvl, int16_t value, const char *namefmt, ...) __printflike(3, 4);
void nvlist_addf_uint16(nvlist_t *nvl, uint16_t value, const char *namefmt, ...) __printflike(3, 4);
void nvlist_addf_int32(nvlist_t *nvl, int32_t value, const char *namefmt, ...) __printflike(3, 4);
void nvlist_addf_uint32(nvlist_t *nvl, uint32_t value, const char *namefmt, ...) __printflike(3, 4);
void nvlist_addf_int64(nvlist_t *nvl, int64_t value, const char *namefmt, ...) __printflike(3, 4);
void nvlist_addf_uint64(nvlist_t *nvl, uint64_t value, const char *namefmt, ...) __printflike(3, 4);
void nvlist_addf_string(nvlist_t *nvl, const char *value, const char *namefmt, ...) __printflike(3, 4);
void nvlist_addf_nvlist(nvlist_t *nvl, const nvlist_t *value, const char *namefmt, ...) __printflike(3, 4);
void nvlist_addf_descriptor(nvlist_t *nvl, int value, const char *namefmt, ...) __printflike(3, 4);
void nvlist_addf_bool_array(nvlist_t *nvl, const bool *value, size_t nitems, const char *namefmt, ...) __printflike(4, 5);
void nvlist_addf_int8_array(nvlist_t *nvl, const int8_t *value, size_t nitems, const char *namefmt, ...) __printflike(4, 5);
void nvlist_addf_uint8_array(nvlist_t *nvl, const uint8_t *value, size_t nitems, const char *namefmt, ...) __printflike(4, 5);
void nvlist_addf_int16_array(nvlist_t *nvl, const int16_t *value, size_t nitems, const char *namefmt, ...) __printflike(4, 5);
void nvlist_addf_uint16_array(nvlist_t *nvl, const uint16_t *value, size_t nitems, const char *namefmt, ...) __printflike(4, 5);
void nvlist_addf_int32_array(nvlist_t *nvl, const int32_t *value, size_t nitems, const char *namefmt, ...) __printflike(4, 5);
void nvlist_addf_uint32_array(nvlist_t *nvl, const uint32_t *value, size_t nitems, const char *namefmt, ...) __printflike(4, 5);
void nvlist_addf_int64_array(nvlist_t *nvl, const int64_t *value, size_t nitems, const char *namefmt, ...) __printflike(4, 5);
void nvlist_addf_uint64_array(nvlist_t *nvl, const uint64_t *value, size_t nitems, const char *namefmt, ...) __printflike(4, 5);
void nvlist_addf_string_array(nvlist_t *nvl, const char * const *value, size_t nitems, const char *namefmt, ...) __printflike(4, 5);
void nvlist_addf_nvlist_array(nvlist_t *nvl, const nvlist_t * const *value, size_t nitems, const char *namefmt, ...) __printflike(4, 5);
void nvlist_addf_descriptor_array(nvlist_t *nvl, const int *value, size_t nitems, const char *namefmt, ...) __printflike(4, 5);

void nvlist_addv_null(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
void nvlist_addv_bool(nvlist_t *nvl, bool value, const char *namefmt, va_list nameap) __printflike(3, 0);
void nvlist_addv_int8(nvlist_t *nvl, int8_t value, const char *namefmt, va_list nameap) __printflike(3, 0);
void nvlist_addv_uint8(nvlist_t *nvl, uint8_t value, const char *namefmt, va_list nameap) __printflike(3, 0);
void nvlist_addv_int16(nvlist_t *nvl, int16_t value, const char *namefmt, va_list nameap) __printflike(3, 0);
void nvlist_addv_uint16(nvlist_t *nvl, uint16_t value, const char *namefmt, va_list nameap) __printflike(3, 0);
void nvlist_addv_int32(nvlist_t *nvl, int32_t value, const char *namefmt, va_list nameap) __printflike(3, 0);
void nvlist_addv_uint32(nvlist_t *nvl, uint32_t value, const char *namefmt, va_list nameap) __printflike(3, 0);
void nvlist_addv_int64(nvlist_t *nvl, int64_t value, const char *namefmt, va_list nameap) __printflike(3, 0);
void nvlist_addv_uint64(nvlist_t *nvl, uint64_t value, const char *namefmt, va_list nameap) __printflike(3, 0);
void nvlist_addv_string(nvlist_t *nvl, const char *value, const char *namefmt, va_list nameap) __printflike(3, 0);
void nvlist_addv_nvlist(nvlist_t *nvl, const nvlist_t *value, const char *namefmt, va_list nameap) __printflike(3, 0);
void nvlist_addv_descriptor(nvlist_t *nvl, int value, const char *namefmt, va_list nameap) __printflike(3, 0);
void nvlist_addv_bool_array(nvlist_t *nvl, const bool *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(4, 0);
void nvlist_addv_int8_array(nvlist_t *nvl, const int8_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(4, 0);
void nvlist_addv_uint8_array(nvlist_t *nvl, const uint8_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(4, 0);
void nvlist_addv_int16_array(nvlist_t *nvl, const int16_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(4, 0);
void nvlist_addv_uint16_array(nvlist_t *nvl, const uint16_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(4, 0);
void nvlist_addv_int32_array(nvlist_t *nvl, const int32_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(4, 0);
void nvlist_addv_uint32_array(nvlist_t *nvl, const uint32_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(4, 0);
void nvlist_addv_int64_array(nvlist_t *nvl, const int64_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(4, 0);
void nvlist_addv_uint64_array(nvlist_t *nvl, const uint64_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(4, 0);
void nvlist_addv_string_array(nvlist_t *nvl, const char * const *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(4, 0);
void nvlist_addv_nvlist_array(nvlist_t *nvl, const nvlist_t * const *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(4, 0);
void nvlist_addv_descriptor_array(nvlist_t *nvl, const int *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(4, 0);

/*
 * The nvlist_move functions add the given name/value pair.
 * The functions consumes provided buffer.
 */

void nvlist_move_string(nvlist_t *nvl, const char *name, char *value);
void nvlist_move_nvlist(nvlist_t *nvl, const char *name, nvlist_t *value);
void nvlist_move_descriptor(nvlist_t *nvl, const char *name, int value);
void nvlist_move_bool_array(nvlist_t *nvl, const char *name, bool *value, size_t nitems);
void nvlist_move_int8_array(nvlist_t *nvl, const char *name, int8_t *value, size_t nitems);
void nvlist_move_uint8_array(nvlist_t *nvl, const char *name, uint8_t *value, size_t nitems);
void nvlist_move_int16_array(nvlist_t *nvl, const char *name, int16_t *value, size_t nitems);
void nvlist_move_uint16_array(nvlist_t *nvl, const char *name, uint16_t *value, size_t nitems);
void nvlist_move_int32_array(nvlist_t *nvl, const char *name, int32_t *value, size_t nitems);
void nvlist_move_uint32_array(nvlist_t *nvl, const char *name, uint32_t *value, size_t nitems);
void nvlist_move_int64_array(nvlist_t *nvl, const char *name, int64_t *value, size_t nitems);
void nvlist_move_uint64_array(nvlist_t *nvl, const char *name, uint64_t *value, size_t nitems);
void nvlist_move_string_array(nvlist_t *nvl, const char *name, char **value, size_t nitems);
void nvlist_move_nvlist_array(nvlist_t *nvl, const char *name, nvlist_t **value, size_t nitems);
void nvlist_move_descriptor_array(nvlist_t *nvl, const char *name, int *value, size_t nitems);

void nvlist_movef_string(nvlist_t *nvl, char *value, const char *namefmt, ...) __printflike(3, 4);
void nvlist_movef_nvlist(nvlist_t *nvl, nvlist_t *value, const char *namefmt, ...) __printflike(3, 4);
void nvlist_movef_descriptor(nvlist_t *nvl, int value, const char *namefmt, ...) __printflike(3, 4);
void nvlist_movef_bool_array(nvlist_t *nvl, bool *value, size_t nitems, const char *namefmt, ...) __printflike(4, 5);
void nvlist_movef_int8_array(nvlist_t *nvl, int8_t *value, size_t nitems, const char *namefmt, ...) __printflike(4, 5);
void nvlist_movef_uint8_array(nvlist_t *nvl, uint8_t *value, size_t nitems, const char *namefmt, ...) __printflike(4, 5);
void nvlist_movef_int16_array(nvlist_t *nvl, int16_t *value, size_t nitems, const char *namefmt, ...) __printflike(4, 5);
void nvlist_movef_uint16_array(nvlist_t *nvl, uint16_t *value, size_t nitems, const char *namefmt, ...) __printflike(4, 5);
void nvlist_movef_int32_array(nvlist_t *nvl, int32_t *value, size_t nitems, const char *namefmt, ...) __printflike(4, 5);
void nvlist_movef_uint32_array(nvlist_t *nvl, uint32_t *value, size_t nitems, const char *namefmt, ...) __printflike(4, 5);
void nvlist_movef_int64_array(nvlist_t *nvl, int64_t *value, size_t nitems, const char *namefmt, ...) __printflike(4, 5);
void nvlist_movef_uint64_array(nvlist_t *nvl, uint64_t *value, size_t nitems, const char *namefmt, ...) __printflike(4, 5);
void nvlist_movef_string_array(nvlist_t *nvl, char **value, size_t nitems, const char *namefmt, ...) __printflike(4, 5);
void nvlist_movef_nvlist_array(nvlist_t *nvl, nvlist_t **value, size_t nitems, const char *namefmt, ...) __printflike(4, 5);
void nvlist_movef_descriptor_array(nvlist_t *nvl, int *value, size_t nitems, const char *namefmt, ...) __printflike(4, 5);

void nvlist_movev_string(nvlist_t *nvl, char *value, const char *namefmt, va_list nameap) __printflike(3, 0);
void nvlist_movev_nvlist(nvlist_t *nvl, nvlist_t *value, const char *namefmt, va_list nameap) __printflike(3, 0);
void nvlist_movev_descriptor(nvlist_t *nvl, int value, const char *namefmt, va_list nameap) __printflike(3, 0);
void nvlist_movev_bool_array(nvlist_t *nvl, bool *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(4, 0);
void nvlist_movev_int8_array(nvlist_t *nvl, int8_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(4, 0);
void nvlist_movev_uint8_array(nvlist_t *nvl, uint8_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(4, 0);
void nvlist_movev_int16_array(nvlist_t *nvl, int16_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(4, 0);
void nvlist_movev_uint16_array(nvlist_t *nvl, uint16_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(4, 0);
void nvlist_movev_int32_array(nvlist_t *nvl, int32_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(4, 0);
void nvlist_movev_uint32_array(nvlist_t *nvl, uint32_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(4, 0);
void nvlist_movev_int64_array(nvlist_t *nvl, int64_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(4, 0);
void nvlist_movev_uint64_array(nvlist_t *nvl, uint64_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(4, 0);
void nvlist_movev_string_array(nvlist_t *nvl, char **value, size_t nitems, const char *namefmt, va_list nameap) __printflike(4, 0);
void nvlist_movev_nvlist_array(nvlist_t *nvl, nvlist_t **value, size_t nitems, const char *namefmt, va_list nameap) __printflike(4, 0);
void nvlist_movev_descriptor_array(nvlist_t *nvl, int *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(4, 0);

/*
 * The nvlist_get functions returns value associated with the given name.
 * If it returns a pointer, the pointer represents internal buffer and should
 * not be freed by the caller.
 */

nvpair_t *nvlist_get_nvpair(const nvlist_t *nvl, const char *name);
bool nvlist_get_bool(const nvlist_t *nvl, const char *name);
int8_t nvlist_get_int8(const nvlist_t *nvl, const char *name);
uint8_t nvlist_get_uint8(const nvlist_t *nvl, const char *name);
int16_t nvlist_get_int16(const nvlist_t *nvl, const char *name);
uint16_t nvlist_get_uint16(const nvlist_t *nvl, const char *name);
int32_t nvlist_get_int32(const nvlist_t *nvl, const char *name);
uint32_t nvlist_get_uint32(const nvlist_t *nvl, const char *name);
int64_t nvlist_get_int64(const nvlist_t *nvl, const char *name);
uint64_t nvlist_get_uint64(const nvlist_t *nvl, const char *name);
const char *nvlist_get_string(const nvlist_t *nvl, const char *name);
const nvlist_t *nvlist_get_nvlist(const nvlist_t *nvl, const char *name);
int nvlist_get_descriptor(const nvlist_t *nvl, const char *name);
const bool *nvlist_get_bool_array(const nvlist_t *nvl, const char *name, size_t *nitemsp);
const int8_t *nvlist_get_int8_array(const nvlist_t *nvl, const char *name, size_t *nitemsp);
const uint8_t *nvlist_get_uint8_array(const nvlist_t *nvl, const char *name, size_t *nitemsp);
const int16_t *nvlist_get_int16_array(const nvlist_t *nvl, const char *name, size_t *nitemsp);
const uint16_t *nvlist_get_uint16_array(const nvlist_t *nvl, const char *name, size_t *nitemsp);
const int32_t *nvlist_get_int32_array(const nvlist_t *nvl, const char *name, size_t *nitemsp);
const uint32_t *nvlist_get_uint32_array(const nvlist_t *nvl, const char *name, size_t *nitemsp);
const int64_t *nvlist_get_int64_array(const nvlist_t *nvl, const char *name, size_t *nitemsp);
const uint64_t *nvlist_get_uint64_array(const nvlist_t *nvl, const char *name, size_t *nitemsp);
const char * const *nvlist_get_string_array(const nvlist_t *nvl, const char *name, size_t *nitemsp);
nvlist_t * const *nvlist_get_nvlist_array(const nvlist_t *nvl, const char *name, size_t *nitemsp);
const int *nvlist_get_descriptor_array(const nvlist_t *nvl, const char *name, size_t *nitemsp);

nvpair_t *nvlist_getf_nvpair(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool nvlist_getf_bool(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int8_t nvlist_getf_int8(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
uint8_t nvlist_getf_uint8(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int16_t nvlist_getf_int16(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
uint16_t nvlist_getf_uint16(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int32_t nvlist_getf_int32(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
uint32_t nvlist_getf_uint32(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int64_t nvlist_getf_int64(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
uint64_t nvlist_getf_uint64(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
const char *nvlist_getf_string(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
const nvlist_t *nvlist_getf_nvlist(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int nvlist_getf_descriptor(const nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
const bool *nvlist_getf_bool_array(const nvlist_t *nvl, size_t *nitemsp, const char *namefmt, ...) __printflike(3, 4);
const int8_t *nvlist_getf_int8_array(const nvlist_t *nvl, size_t *nitemsp, const char *namefmt, ...) __printflike(3, 4);
const uint8_t *nvlist_getf_uint8_array(const nvlist_t *nvl, size_t *nitemsp, const char *namefmt, ...) __printflike(3, 4);
const int16_t *nvlist_getf_int16_array(const nvlist_t *nvl, size_t *nitemsp, const char *namefmt, ...) __printflike(3, 4);
const uint16_t *nvlist_getf_uint16_array(const nvlist_t *nvl, size_t *nitemsp, const char *namefmt, ...) __printflike(3, 4);
const int32_t *nvlist_getf_int32_array(const nvlist_t *nvl, size_t *nitemsp, const char *namefmt, ...) __printflike(3, 4);
const uint32_t *nvlist_getf_uint32_array(const nvlist_t *nvl, size_t *nitemsp, const char *namefmt, ...) __printflike(3, 4);
const int64_t *nvlist_getf_int64_array(const nvlist_t *nvl, size_t *nitemsp, const char *namefmt, ...) __printflike(3, 4);
const uint64_t *nvlist_getf_uint64_array(const nvlist_t *nvl, size_t *nitemsp, const char *namefmt, ...) __printflike(3, 4);
const char * const *nvlist_getf_string_array(const nvlist_t *nvl, size_t *nitemsp, const char *namefmt, ...) __printflike(3, 4);
nvlist_t * const *nvlist_getf_nvlist_array(const nvlist_t *nvl, size_t *nitemsp, const char *namefmt, ...) __printflike(3, 4);
const int *nvlist_getf_descriptor_array(const nvlist_t *nvl, size_t *nitemsp, const char *namefmt, ...) __printflike(3, 4);

nvpair_t *nvlist_getv_nvpair(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool nvlist_getv_bool(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int8_t nvlist_getv_int8(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
uint8_t nvlist_getv_uint8(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int16_t nvlist_getv_int16(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
uint16_t nvlist_getv_uint16(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int32_t nvlist_getv_int32(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
uint32_t nvlist_getv_uint32(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int64_t nvlist_getv_int64(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
uint64_t nvlist_getv_uint64(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
const char *nvlist_getv_string(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
const nvlist_t *nvlist_getv_nvlist(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int nvlist_getv_descriptor(const nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
const bool *nvlist_getv_bool_array(const nvlist_t *nvl, size_t *nitemsp, const char *namefmt, va_list nameap) __printflike(3, 0);
const int8_t *nvlist_getv_int8_array(const nvlist_t *nvl, size_t *nitemsp, const char *namefmt, va_list nameap) __printflike(3, 0);
const uint8_t *nvlist_getv_uint8_array(const nvlist_t *nvl, size_t *nitemsp, const char *namefmt, va_list nameap) __printflike(3, 0);
const int16_t *nvlist_getv_int16_array(const nvlist_t *nvl, size_t *nitemsp, const char *namefmt, va_list nameap) __printflike(3, 0);
const uint16_t *nvlist_getv_uint16_array(const nvlist_t *nvl, size_t *nitemsp, const char *namefmt, va_list nameap) __printflike(3, 0);
const int32_t *nvlist_getv_int32_array(const nvlist_t *nvl, size_t *nitemsp, const char *namefmt, va_list nameap) __printflike(3, 0);
const uint32_t *nvlist_getv_uint32_array(const nvlist_t *nvl, size_t *nitemsp, const char *namefmt, va_list nameap) __printflike(3, 0);
const int64_t *nvlist_getv_int64_array(const nvlist_t *nvl, size_t *nitemsp, const char *namefmt, va_list nameap) __printflike(3, 0);
const uint64_t *nvlist_getv_uint64_array(const nvlist_t *nvl, size_t *nitemsp, const char *namefmt, va_list nameap) __printflike(3, 0);
const char * const *nvlist_getv_string_array(const nvlist_t *nvl, size_t *nitemsp, const char *namefmt, va_list nameap) __printflike(3, 0);
nvlist_t * const *nvlist_getv_nvlist_array(const nvlist_t *nvl, size_t *nitemsp, const char *namefmt, va_list nameap) __printflike(3, 0);
const int *nvlist_getv_descriptor_array(const nvlist_t *nvl, size_t *nitemsp, const char *namefmt, va_list nameap) __printflike(3, 0);

/*
 * The nvlist_take functions returns value associated with the given name and
 * remove the given entry from the nvlist.
 * The caller is responsible for freeing received data.
 */

nvpair_t *nvlist_take_nvpair(nvlist_t *nvl, const char *name);
bool nvlist_take_bool(nvlist_t *nvl, const char *name);
int8_t nvlist_take_int8(nvlist_t *nvl, const char *name);
uint8_t nvlist_take_uint8(nvlist_t *nvl, const char *name);
int16_t nvlist_take_int16(nvlist_t *nvl, const char *name);
uint16_t nvlist_take_uint16(nvlist_t *nvl, const char *name);
int32_t nvlist_take_int32(nvlist_t *nvl, const char *name);
uint32_t nvlist_take_uint32(nvlist_t *nvl, const char *name);
int64_t nvlist_take_int64(nvlist_t *nvl, const char *name);
uint64_t nvlist_take_uint64(nvlist_t *nvl, const char *name);
char *nvlist_take_string(nvlist_t *nvl, const char *name);
nvlist_t *nvlist_take_nvlist(nvlist_t *nvl, const char *name);
int nvlist_take_descriptor(nvlist_t *nvl, const char *name);
bool *nvlist_take_bool_array(nvlist_t *nvl, const char *name, size_t *nitemsp);
int8_t *nvlist_take_int8_array(nvlist_t *nvl, const char *name, size_t *nitemsp);
uint8_t *nvlist_take_uint8_array(nvlist_t *nvl, const char *name, size_t *nitemsp);
int16_t *nvlist_take_int16_array(nvlist_t *nvl, const char *name, size_t *nitemsp);
uint16_t *nvlist_take_uint16_array(nvlist_t *nvl, const char *name, size_t *nitemsp);
int32_t *nvlist_take_int32_array(nvlist_t *nvl, const char *name, size_t *nitemsp);
uint32_t *nvlist_take_uint32_array(nvlist_t *nvl, const char *name, size_t *nitemsp);
int64_t *nvlist_take_int64_array(nvlist_t *nvl, const char *name, size_t *nitemsp);
uint64_t *nvlist_take_uint64_array(nvlist_t *nvl, const char *name, size_t *nitemsp);
char **nvlist_take_string_array(nvlist_t *nvl, const char *name, size_t *nitemsp);
nvlist_t **nvlist_take_nvlist_array(nvlist_t *nvl, const char *name, size_t *nitemsp);
int *nvlist_take_descriptor_array(nvlist_t *nvl, const char *name, size_t *nitemsp);

nvpair_t *nvlist_takef_nvpair(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool nvlist_takef_bool(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int8_t nvlist_takef_int8(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
uint8_t nvlist_takef_uint8(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int16_t nvlist_takef_int16(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
uint16_t nvlist_takef_uint16(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int32_t nvlist_takef_int32(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
uint32_t nvlist_takef_uint32(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int64_t nvlist_takef_int64(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
uint64_t nvlist_takef_uint64(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
char *nvlist_takef_string(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
nvlist_t *nvlist_takef_nvlist(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int nvlist_takef_descriptor(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
bool *nvlist_takef_bool_array(nvlist_t *nvl, size_t *nitemsp, const char *namefmt, ...) __printflike(3, 4);
int8_t *nvlist_takef_int8_array(nvlist_t *nvl, size_t *nitemsp, const char *namefmt, ...) __printflike(3, 4);
uint8_t *nvlist_takef_uint8_array(nvlist_t *nvl, size_t *nitemsp, const char *namefmt, ...) __printflike(3, 4);
int16_t *nvlist_takef_int16_array(nvlist_t *nvl, size_t *nitemsp, const char *namefmt, ...) __printflike(3, 4);
uint16_t *nvlist_takef_uint16_array(nvlist_t *nvl, size_t *nitemsp, const char *namefmt, ...) __printflike(3, 4);
int32_t *nvlist_takef_int32_array(nvlist_t *nvl, size_t *nitemsp, const char *namefmt, ...) __printflike(3, 4);
uint32_t *nvlist_takef_uint32_array(nvlist_t *nvl, size_t *nitemsp, const char *namefmt, ...) __printflike(3, 4);
int64_t *nvlist_takef_int64_array(nvlist_t *nvl, size_t *nitemsp, const char *namefmt, ...) __printflike(3, 4);
uint64_t *nvlist_takef_uint64_array(nvlist_t *nvl, size_t *nitemsp, const char *namefmt, ...) __printflike(3, 4);
char **nvlist_takef_string_array(nvlist_t *nvl, size_t *nitemsp, const char *namefmt, ...) __printflike(3, 4);
nvlist_t **nvlist_takef_nvlist_array(nvlist_t *nvl, size_t *nitemsp, const char *namefmt, ...) __printflike(3, 4);
int *nvlist_takef_descriptor_array(nvlist_t *nvl, size_t *nitemsp, const char *namefmt, ...) __printflike(3, 4);

nvpair_t *nvlist_takev_nvpair(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool nvlist_takev_bool(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int8_t nvlist_takev_int8(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
uint8_t nvlist_takev_uint8(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int16_t nvlist_takev_int16(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
uint16_t nvlist_takev_uint16(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int32_t nvlist_takev_int32(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
uint32_t nvlist_takev_uint32(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int64_t nvlist_takev_int64(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
uint64_t nvlist_takev_uint64(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
char *nvlist_takev_string(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
nvlist_t *nvlist_takev_nvlist(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int nvlist_takev_descriptor(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
bool *nvlist_takev_bool_array(nvlist_t *nvl, size_t *nitemsp, const char *namefmt, va_list nameap) __printflike(3, 0);
int8_t *nvlist_takev_int8_array(nvlist_t *nvl, size_t *nitemsp, const char *namefmt, va_list nameap) __printflike(3, 0);
uint8_t *nvlist_takev_uint8_array(nvlist_t *nvl, size_t *nitemsp, const char *namefmt, va_list nameap) __printflike(3, 0);
int16_t *nvlist_takev_int16_array(nvlist_t *nvl, size_t *nitemsp, const char *namefmt, va_list nameap) __printflike(3, 0);
uint16_t *nvlist_takev_uint16_array(nvlist_t *nvl, size_t *nitemsp, const char *namefmt, va_list nameap) __printflike(3, 0);
int32_t *nvlist_takev_int32_array(nvlist_t *nvl, size_t *nitemsp, const char *namefmt, va_list nameap) __printflike(3, 0);
uint32_t *nvlist_takev_uint32_array(nvlist_t *nvl, size_t *nitemsp, const char *namefmt, va_list nameap) __printflike(3, 0);
int64_t *nvlist_takev_int64_array(nvlist_t *nvl, size_t *nitemsp, const char *namefmt, va_list nameap) __printflike(3, 0);
uint64_t *nvlist_takev_uint64_array(nvlist_t *nvl, size_t *nitemsp, const char *namefmt, va_list nameap) __printflike(3, 0);
char **nvlist_takev_string_array(nvlist_t *nvl, size_t *nitemsp, const char *namefmt, va_list nameap) __printflike(3, 0);
nvlist_t **nvlist_takev_nvlist_array(nvlist_t *nvl, size_t *nitemsp, const char *namefmt, va_list nameap) __printflike(3, 0);
int *nvlist_takev_descriptor_array(nvlist_t *nvl, size_t *nitemsp, const char *namefmt, va_list nameap) __printflike(3, 0);

/* Function removes the given nvpair from the nvlist. */
void nvlist_remove_nvpair(nvlist_t *nvl, nvpair_t *nvp);

/*
 * The nvlist_free functions removes the given name/value pair from the nvlist
 * and frees memory associated with it.
 */

int nvlist_free(nvlist_t *nvl, const char *name);
int nvlist_free_null(nvlist_t *nvl, const char *name);
int nvlist_free_bool(nvlist_t *nvl, const char *name);
int nvlist_free_int8(nvlist_t *nvl, const char *name);
int nvlist_free_uint8(nvlist_t *nvl, const char *name);
int nvlist_free_int16(nvlist_t *nvl, const char *name);
int nvlist_free_uint16(nvlist_t *nvl, const char *name);
int nvlist_free_int32(nvlist_t *nvl, const char *name);
int nvlist_free_uint32(nvlist_t *nvl, const char *name);
int nvlist_free_int64(nvlist_t *nvl, const char *name);
int nvlist_free_uint64(nvlist_t *nvl, const char *name);
int nvlist_free_string(nvlist_t *nvl, const char *name);
int nvlist_free_nvlist(nvlist_t *nvl, const char *name);
int nvlist_free_descriptor(nvlist_t *nvl, const char *name);
int nvlist_free_bool_array(nvlist_t *nvl, const char *name);
int nvlist_free_int8_array(nvlist_t *nvl, const char *name);
int nvlist_free_uint8_array(nvlist_t *nvl, const char *name);
int nvlist_free_int16_array(nvlist_t *nvl, const char *name);
int nvlist_free_uint16_array(nvlist_t *nvl, const char *name);
int nvlist_free_int32_array(nvlist_t *nvl, const char *name);
int nvlist_free_uint32_array(nvlist_t *nvl, const char *name);
int nvlist_free_int64_array(nvlist_t *nvl, const char *name);
int nvlist_free_uint64_array(nvlist_t *nvl, const char *name);
int nvlist_free_string_array(nvlist_t *nvl, const char *name);
int nvlist_free_nvlist_array(nvlist_t *nvl, const char *name);
int nvlist_free_descriptor_array(nvlist_t *nvl, const char *name);

int nvlist_freef(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int nvlist_freef_null(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int nvlist_freef_bool(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int nvlist_freef_int8(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int nvlist_freef_uint8(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int nvlist_freef_int16(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int nvlist_freef_uint16(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int nvlist_freef_int32(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int nvlist_freef_uint32(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int nvlist_freef_int64(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int nvlist_freef_uint64(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int nvlist_freef_string(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int nvlist_freef_nvlist(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int nvlist_freef_descriptor(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int nvlist_freef_bool_array(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int nvlist_freef_int8_array(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int nvlist_freef_uint8_array(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int nvlist_freef_int16_array(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int nvlist_freef_uint16_array(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int nvlist_freef_int32_array(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int nvlist_freef_uint32_array(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int nvlist_freef_int64_array(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int nvlist_freef_uint64_array(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int nvlist_freef_string_array(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int nvlist_freef_nvlist_array(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);
int nvlist_freef_descriptor_array(nvlist_t *nvl, const char *namefmt, ...) __printflike(2, 3);

int nvlist_freev(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int nvlist_freev_null(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int nvlist_freev_bool(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int nvlist_freev_int8(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int nvlist_freev_uint8(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int nvlist_freev_int16(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int nvlist_freev_uint16(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int nvlist_freev_int32(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int nvlist_freev_uint32(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int nvlist_freev_int64(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int nvlist_freev_uint64(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int nvlist_freev_string(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int nvlist_freev_nvlist(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int nvlist_freev_descriptor(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int nvlist_freev_bool_array(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int nvlist_freev_int8_array(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int nvlist_freev_uint8_array(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int nvlist_freev_int16_array(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int nvlist_freev_uint16_array(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int nvlist_freev_int32_array(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int nvlist_freev_uint32_array(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int nvlist_freev_int64_array(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int nvlist_freev_uint64_array(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int nvlist_freev_string_array(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int nvlist_freev_nvlist_array(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);
int nvlist_freev_descriptor_array(nvlist_t *nvl, const char *namefmt, va_list nameap) __printflike(2, 0);

void nvlist_free_nvpair(nvlist_t *nvl, nvpair_t *nvp);

int nvpair_type(const nvpair_t *nvp);
const char *nvpair_name(const nvpair_t *nvp);

nvpair_t *nvpair_clone(const nvpair_t *nvp);

nvpair_t *nvpair_create_null(const char *name);
nvpair_t *nvpair_create_bool(const char *name, bool value);
nvpair_t *nvpair_create_int8(const char *name, int8_t value);
nvpair_t *nvpair_create_uint8(const char *name, uint8_t value);
nvpair_t *nvpair_create_int16(const char *name, int16_t value);
nvpair_t *nvpair_create_uint16(const char *name, uint16_t value);
nvpair_t *nvpair_create_int32(const char *name, int32_t value);
nvpair_t *nvpair_create_uint32(const char *name, uint32_t value);
nvpair_t *nvpair_create_int64(const char *name, int64_t value);
nvpair_t *nvpair_create_uint64(const char *name, uint64_t value);
nvpair_t *nvpair_create_string(const char *name, const char *value);
nvpair_t *nvpair_create_stringf(const char *name, const char *valuefmt, ...) __printflike(2, 3);
nvpair_t *nvpair_create_stringv(const char *name, const char *valuefmt, va_list valueap) __printflike(2, 0);
nvpair_t *nvpair_create_nvlist(const char *name, const nvlist_t *value);
nvpair_t *nvpair_create_descriptor(const char *name, int value);
nvpair_t *nvpair_create_bool_array(const char *name, const bool *value, size_t nitems);
nvpair_t *nvpair_create_int8_array(const char *name, const int8_t *value, size_t nitems);
nvpair_t *nvpair_create_uint8_array(const char *name, const uint8_t *value, size_t nitems);
nvpair_t *nvpair_create_int16_array(const char *name, const int16_t *value, size_t nitems);
nvpair_t *nvpair_create_uint16_array(const char *name, const uint16_t *value, size_t nitems);
nvpair_t *nvpair_create_int32_array(const char *name, const int32_t *value, size_t nitems);
nvpair_t *nvpair_create_uint32_array(const char *name, const uint32_t *value, size_t nitems);
nvpair_t *nvpair_create_int64_array(const char *name, const int64_t *value, size_t nitems);
nvpair_t *nvpair_create_uint64_array(const char *name, const uint64_t *value, size_t nitems);
nvpair_t *nvpair_create_string_array(const char *name, const char * const *value, size_t nitems);
nvpair_t *nvpair_create_nvlist_array(const char *name, const nvlist_t * const *value, size_t nitems);
nvpair_t *nvpair_create_descriptor_array(const char *name, const int *value, size_t nitems);

nvpair_t *nvpair_createf_null(const char *namefmt, ...) __printflike(1, 2);
nvpair_t *nvpair_createf_bool(bool value, const char *namefmt, ...) __printflike(2, 3);
nvpair_t *nvpair_createf_int8(int8_t value, const char *namefmt, ...) __printflike(2, 3);
nvpair_t *nvpair_createf_uint8(uint8_t value, const char *namefmt, ...) __printflike(2, 3);
nvpair_t *nvpair_createf_int16(int16_t value, const char *namefmt, ...) __printflike(2, 3);
nvpair_t *nvpair_createf_uint16(uint16_t value, const char *namefmt, ...) __printflike(2, 3);
nvpair_t *nvpair_createf_int32(int32_t value, const char *namefmt, ...) __printflike(2, 3);
nvpair_t *nvpair_createf_uint32(uint32_t value, const char *namefmt, ...) __printflike(2, 3);
nvpair_t *nvpair_createf_int64(int64_t value, const char *namefmt, ...) __printflike(2, 3);
nvpair_t *nvpair_createf_uint64(uint64_t value, const char *namefmt, ...) __printflike(2, 3);
nvpair_t *nvpair_createf_string(const char *value, const char *namefmt, ...) __printflike(2, 3);
nvpair_t *nvpair_createf_nvlist(const nvlist_t *value, const char *namefmt, ...) __printflike(2, 3);
nvpair_t *nvpair_createf_descriptor(int value, const char *namefmt, ...) __printflike(2, 3);
nvpair_t *nvpair_createf_bool_array(const bool *value, size_t nitems, const char *namefmt, ...) __printflike(3, 4);
nvpair_t *nvpair_createf_int8_array(const int8_t *value, size_t nitems, const char *namefmt, ...) __printflike(3, 4);
nvpair_t *nvpair_createf_uint8_array(const uint8_t *value, size_t nitems, const char *namefmt, ...) __printflike(3, 4);
nvpair_t *nvpair_createf_int16_array(const int16_t *value, size_t nitems, const char *namefmt, ...) __printflike(3, 4);
nvpair_t *nvpair_createf_uint16_array(const uint16_t *value, size_t nitems, const char *namefmt, ...) __printflike(3, 4);
nvpair_t *nvpair_createf_int32_array(const int32_t *value, size_t nitems, const char *namefmt, ...) __printflike(3, 4);
nvpair_t *nvpair_createf_uint32_array(const uint32_t *value, size_t nitems, const char *namefmt, ...) __printflike(3, 4);
nvpair_t *nvpair_createf_int64_array(const int64_t *value, size_t nitems, const char *namefmt, ...) __printflike(3, 4);
nvpair_t *nvpair_createf_uint64_array(const uint64_t *value, size_t nitems, const char *namefmt, ...) __printflike(3, 4);
nvpair_t *nvpair_createf_string_array(const char * const *value, size_t nitems, const char *namefmt, ...) __printflike(3, 4);
nvpair_t *nvpair_createf_nvlist_array(const nvlist_t * const *value, size_t nitems, const char *namefmt, ...) __printflike(3, 4);
nvpair_t *nvpair_createf_descriptor_array(const int *value, size_t nitems, const char *namefmt, ...) __printflike(3, 4);

nvpair_t *nvpair_createv_null(const char *namefmt, va_list nameap) __printflike(1, 0);
nvpair_t *nvpair_createv_bool(bool value, const char *namefmt, va_list nameap) __printflike(2, 0);
nvpair_t *nvpair_createv_int8(int8_t value, const char *namefmt, va_list nameap) __printflike(2, 0);
nvpair_t *nvpair_createv_uint8(uint8_t value, const char *namefmt, va_list nameap) __printflike(2, 0);
nvpair_t *nvpair_createv_int16(int16_t value, const char *namefmt, va_list nameap) __printflike(2, 0);
nvpair_t *nvpair_createv_uint16(uint16_t value, const char *namefmt, va_list nameap) __printflike(2, 0);
nvpair_t *nvpair_createv_int32(int32_t value, const char *namefmt, va_list nameap) __printflike(2, 0);
nvpair_t *nvpair_createv_uint32(uint32_t value, const char *namefmt, va_list nameap) __printflike(2, 0);
nvpair_t *nvpair_createv_int64(int64_t value, const char *namefmt, va_list nameap) __printflike(2, 0);
nvpair_t *nvpair_createv_uint64(uint64_t value, const char *namefmt, va_list nameap) __printflike(2, 0);
nvpair_t *nvpair_createv_string(const char *value, const char *namefmt, va_list nameap) __printflike(2, 0);
nvpair_t *nvpair_createv_nvlist(const nvlist_t *value, const char *namefmt, va_list nameap) __printflike(2, 0);
nvpair_t *nvpair_createv_descriptor(int value, const char *namefmt, va_list nameap) __printflike(2, 0);
nvpair_t *nvpair_createv_bool_array(const bool *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(3, 0);
nvpair_t *nvpair_createv_int8_array(const int8_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(3, 0);
nvpair_t *nvpair_createv_uint8_array(const uint8_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(3, 0);
nvpair_t *nvpair_createv_int16_array(const int16_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(3, 0);
nvpair_t *nvpair_createv_uint16_array(const uint16_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(3, 0);
nvpair_t *nvpair_createv_int32_array(const int32_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(3, 0);
nvpair_t *nvpair_createv_uint32_array(const uint32_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(3, 0);
nvpair_t *nvpair_createv_int64_array(const int64_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(3, 0);
nvpair_t *nvpair_createv_uint64_array(const uint64_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(3, 0);
nvpair_t *nvpair_createv_string_array(const char * const *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(3, 0);
nvpair_t *nvpair_createv_nvlist_array(const nvlist_t * const *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(3, 0);
nvpair_t *nvpair_createv_descriptor_array(const int *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(3, 0);

nvpair_t *nvpair_move_string(const char *name, char *value);
nvpair_t *nvpair_move_nvlist(const char *name, nvlist_t *value);
nvpair_t *nvpair_move_descriptor(const char *name, int value);
nvpair_t *nvpair_move_bool_array(const char *name, bool *value, size_t nitems);
nvpair_t *nvpair_move_int8_array(const char *name, int8_t *value, size_t nitems);
nvpair_t *nvpair_move_uint8_array(const char *name, uint8_t *value, size_t nitems);
nvpair_t *nvpair_move_int16_array(const char *name, int16_t *value, size_t nitems);
nvpair_t *nvpair_move_uint16_array(const char *name, uint16_t *value, size_t nitems);
nvpair_t *nvpair_move_int32_array(const char *name, int32_t *value, size_t nitems);
nvpair_t *nvpair_move_uint32_array(const char *name, uint32_t *value, size_t nitems);
nvpair_t *nvpair_move_int64_array(const char *name, int64_t *value, size_t nitems);
nvpair_t *nvpair_move_uint64_array(const char *name, uint64_t *value, size_t nitems);
nvpair_t *nvpair_move_string_array(const char *name, char **value, size_t nitems);
nvpair_t *nvpair_move_nvlist_array(const char *name, nvlist_t **value, size_t nitems);
nvpair_t *nvpair_move_descriptor_array(const char *name, int *value, size_t nitems);

nvpair_t *nvpair_movef_string(char *value, const char *namefmt, ...) __printflike(2, 3);
nvpair_t *nvpair_movef_nvlist(nvlist_t *value, const char *namefmt, ...) __printflike(2, 3);
nvpair_t *nvpair_movef_descriptor(int value, const char *namefmt, ...) __printflike(2, 3);
nvpair_t *nvpair_movef_bool_array(bool *value, size_t nitems, const char *namefmt, ...) __printflike(3, 4);
nvpair_t *nvpair_movef_int8_array(int8_t *value, size_t nitems, const char *namefmt, ...) __printflike(3, 4);
nvpair_t *nvpair_movef_uint8_array(uint8_t *value, size_t nitems, const char *namefmt, ...) __printflike(3, 4);
nvpair_t *nvpair_movef_int16_array(int16_t *value, size_t nitems, const char *namefmt, ...) __printflike(3, 4);
nvpair_t *nvpair_movef_uint16_array(uint16_t *value, size_t nitems, const char *namefmt, ...) __printflike(3, 4);
nvpair_t *nvpair_movef_int32_array(int32_t *value, size_t nitems, const char *namefmt, ...) __printflike(3, 4);
nvpair_t *nvpair_movef_uint32_array(uint32_t *value, size_t nitems, const char *namefmt, ...) __printflike(3, 4);
nvpair_t *nvpair_movef_int64_array(int64_t *value, size_t nitems, const char *namefmt, ...) __printflike(3, 4);
nvpair_t *nvpair_movef_uint64_array(uint64_t *value, size_t nitems, const char *namefmt, ...) __printflike(3, 4);
nvpair_t *nvpair_movef_string_array(char **value, size_t nitems, const char *namefmt, ...) __printflike(3, 4);
nvpair_t *nvpair_movef_nvlist_array(nvlist_t **value, size_t nitems, const char *namefmt, ...) __printflike(3, 4);
nvpair_t *nvpair_movef_descriptor_array(int *value, size_t nitems, const char *namefmt, ...) __printflike(3, 4);

nvpair_t *nvpair_movev_string(char *value, const char *namefmt, va_list nameap) __printflike(2, 0);
nvpair_t *nvpair_movev_nvlist(nvlist_t *value, const char *namefmt, va_list nameap) __printflike(2, 0);
nvpair_t *nvpair_movev_descriptor(int value, const char *namefmt, va_list nameap) __printflike(2, 0);
nvpair_t *nvpair_movev_bool_array(bool *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(3, 0);
nvpair_t *nvpair_movev_int8_array(int8_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(3, 0);
nvpair_t *nvpair_movev_uint8_array(uint8_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(3, 0);
nvpair_t *nvpair_movev_int16_array(int16_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(3, 0);
nvpair_t *nvpair_movev_uint16_array(uint16_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(3, 0);
nvpair_t *nvpair_movev_int32_array(int32_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(3, 0);
nvpair_t *nvpair_movev_uint32_array(uint32_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(3, 0);
nvpair_t *nvpair_movev_int64_array(int64_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(3, 0);
nvpair_t *nvpair_movev_uint64_array(uint64_t *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(3, 0);
nvpair_t *nvpair_movev_string_array(char **value, size_t nitems, const char *namefmt, va_list nameap) __printflike(3, 0);
nvpair_t *nvpair_movev_nvlist_array(nvlist_t **value, size_t nitems, const char *namefmt, va_list nameap) __printflike(3, 0);
nvpair_t *nvpair_movev_descriptor_array(int *value, size_t nitems, const char *namefmt, va_list nameap) __printflike(3, 0);

bool nvpair_get_bool(const nvpair_t *nvp);
int8_t nvpair_get_int8(const nvpair_t *nvp);
uint8_t nvpair_get_uint8(const nvpair_t *nvp);
int16_t nvpair_get_int16(const nvpair_t *nvp);
uint16_t nvpair_get_uint16(const nvpair_t *nvp);
int32_t nvpair_get_int32(const nvpair_t *nvp);
uint32_t nvpair_get_uint32(const nvpair_t *nvp);
int64_t nvpair_get_int64(const nvpair_t *nvp);
uint64_t nvpair_get_uint64(const nvpair_t *nvp);
const char *nvpair_get_string(const nvpair_t *nvp);
const nvlist_t *nvpair_get_nvlist(const nvpair_t *nvp);
int nvpair_get_descriptor(const nvpair_t *nvp);
const bool *nvpair_get_bool_array(const nvpair_t *nvp, size_t *nitemsp);
const int8_t *nvpair_get_int8_array(const nvpair_t *nvp, size_t *nitemsp);
const uint8_t *nvpair_get_uint8_array(const nvpair_t *nvp, size_t *nitemsp);
const int16_t *nvpair_get_int16_array(const nvpair_t *nvp, size_t *nitemsp);
const uint16_t *nvpair_get_uint16_array(const nvpair_t *nvp, size_t *nitemsp);
const int32_t *nvpair_get_int32_array(const nvpair_t *nvp, size_t *nitemsp);
const uint32_t *nvpair_get_uint32_array(const nvpair_t *nvp, size_t *nitemsp);
const int64_t *nvpair_get_int64_array(const nvpair_t *nvp, size_t *nitemsp);
const uint64_t *nvpair_get_uint64_array(const nvpair_t *nvp, size_t *nitemsp);
const char * const *nvpair_get_string_array(const nvpair_t *nvp, size_t *nitemsp);
nvlist_t * const *nvpair_get_nvlist_array(const nvpair_t *nvp, size_t *nitemsp);
const int *nvpair_get_descriptor_array(const nvpair_t *nvp, size_t *nitemsp);

void nvpair_free(nvpair_t *nvp);

#endif	/* !_NV_H_ */
