/*-
 * Copyright (c) 2009-2013 The FreeBSD Foundation
 * All rights reserved.
 *
 * This software was developed by Pawel Jakub Dawidek under sponsorship from
 * the FreeBSD Foundation.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/param.h>
#include <sys/endian.h>
#include <sys/queue.h>

#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifdef HAVE_PJDLOG
#include <pjdlog.h>
#endif

#include "nv.h"
#include "nvlist_impl.h"
#include "nvpair_impl.h"

#ifndef	HAVE_PJDLOG
#include <assert.h>
#define	PJDLOG_ASSERT(...)		assert(__VA_ARGS__)
#define	PJDLOG_RASSERT(expr, ...)	assert(expr)
#define	PJDLOG_ABORT(...)		abort()
#endif

#define	NV_TYPE_FIRST		NV_TYPE_NULL
#define	NV_TYPE_LAST		NV_TYPE_DESCRIPTOR_ARRAY

#define	NV_FLAG_BIG_ENDIAN		0x80

#define	NVPAIR_MAGIC	0x6e7670	/* "nvp" */
struct nvpair {
	int		 nvp_magic;
	char		*nvp_name;
	int		 nvp_type;
	uint64_t	 nvp_data;
	size_t		 nvp_datasize;
	size_t		 nvp_nitems;	/* Used only for array types. */
	nvlist_t	*nvp_list;	/* Used for sanity checks. */
	TAILQ_ENTRY(nvpair) nvp_next;
};

#define	NVPAIR_ASSERT(nvp)	do {					\
	PJDLOG_ASSERT((nvp) != NULL);					\
	PJDLOG_ASSERT((nvp)->nvp_magic == NVPAIR_MAGIC);		\
} while (0)

struct nvpair_header {
	uint8_t		nvph_type;
	uint16_t	nvph_namesize;
	uint64_t	nvph_datasize;
} __packed;

static bool
fd_is_valid(int fd)
{

	return (fcntl(fd, F_GETFL) != -1 || errno != EBADF);
}

void
nvpair_assert(const nvpair_t *nvp)
{

	NVPAIR_ASSERT(nvp);
}

const nvlist_t *
nvpair_nvlist(const nvpair_t *nvp)
{

	NVPAIR_ASSERT(nvp);

	return (nvp->nvp_list);
}

nvpair_t *
nvpair_next(const nvpair_t *nvp)
{

	NVPAIR_ASSERT(nvp);
	PJDLOG_ASSERT(nvp->nvp_list != NULL);

	return (TAILQ_NEXT(nvp, nvp_next));
}

nvpair_t *
nvpair_prev(const nvpair_t *nvp)
{

	NVPAIR_ASSERT(nvp);
	PJDLOG_ASSERT(nvp->nvp_list != NULL);

	return (TAILQ_PREV(nvp, nvl_head, nvp_next));
}

void
nvpair_insert(struct nvl_head *head, nvpair_t *nvp, nvlist_t *nvl)
{

	NVPAIR_ASSERT(nvp);
	PJDLOG_ASSERT(nvp->nvp_list == NULL);

	TAILQ_INSERT_TAIL(head, nvp, nvp_next);
	nvp->nvp_list = nvl;
}

void
nvpair_remove(struct nvl_head *head, nvpair_t *nvp, const nvlist_t *nvl)
{

	NVPAIR_ASSERT(nvp);
	PJDLOG_ASSERT(nvp->nvp_list == nvl);

	TAILQ_REMOVE(head, nvp, nvp_next);
	nvp->nvp_list = NULL;
}

nvpair_t *
nvpair_clone(const nvpair_t *nvp)
{
	nvpair_t *newnvp;
	const char *name;
	const void *data;
	size_t nitems;

	NVPAIR_ASSERT(nvp);

	name = nvpair_name(nvp);

	switch (nvpair_type(nvp)) {
	case NV_TYPE_NULL:
		newnvp = nvpair_create_null(name);
		break;
	case NV_TYPE_BOOL:
		newnvp = nvpair_create_bool(name, nvpair_get_bool(nvp));
		break;
	case NV_TYPE_INT8:
		newnvp = nvpair_create_int8(name, nvpair_get_int8(nvp));
		break;
	case NV_TYPE_UINT8:
		newnvp = nvpair_create_uint8(name, nvpair_get_uint8(nvp));
		break;
	case NV_TYPE_INT16:
		newnvp = nvpair_create_int16(name, nvpair_get_int16(nvp));
		break;
	case NV_TYPE_UINT16:
		newnvp = nvpair_create_uint16(name, nvpair_get_uint16(nvp));
		break;
	case NV_TYPE_INT32:
		newnvp = nvpair_create_int32(name, nvpair_get_int32(nvp));
		break;
	case NV_TYPE_UINT32:
		newnvp = nvpair_create_uint32(name, nvpair_get_uint32(nvp));
		break;
	case NV_TYPE_INT64:
		newnvp = nvpair_create_int64(name, nvpair_get_int64(nvp));
		break;
	case NV_TYPE_UINT64:
		newnvp = nvpair_create_uint64(name, nvpair_get_uint64(nvp));
		break;
	case NV_TYPE_STRING:
		newnvp = nvpair_create_string(name, nvpair_get_string(nvp));
		break;
	case NV_TYPE_NVLIST:
		newnvp = nvpair_create_nvlist(name, nvpair_get_nvlist(nvp));
		break;
	case NV_TYPE_DESCRIPTOR:
		newnvp = nvpair_create_descriptor(name,
		    nvpair_get_descriptor(nvp));
		break;
	case NV_TYPE_BOOL_ARRAY:
		data = nvpair_get_bool_array(nvp, &nitems);
		newnvp = nvpair_create_bool_array(name, data, nitems);
		break;
	case NV_TYPE_INT8_ARRAY:
		data = nvpair_get_int8_array(nvp, &nitems);
		newnvp = nvpair_create_int8_array(name, data, nitems);
		break;
	case NV_TYPE_UINT8_ARRAY:
		data = nvpair_get_uint8_array(nvp, &nitems);
		newnvp = nvpair_create_uint8_array(name, data, nitems);
		break;
	case NV_TYPE_INT16_ARRAY:
		data = nvpair_get_int16_array(nvp, &nitems);
		newnvp = nvpair_create_int16_array(name, data, nitems);
		break;
	case NV_TYPE_UINT16_ARRAY:
		data = nvpair_get_uint16_array(nvp, &nitems);
		newnvp = nvpair_create_uint16_array(name, data, nitems);
		break;
	case NV_TYPE_INT32_ARRAY:
		data = nvpair_get_int32_array(nvp, &nitems);
		newnvp = nvpair_create_int32_array(name, data, nitems);
		break;
	case NV_TYPE_UINT32_ARRAY:
		data = nvpair_get_uint32_array(nvp, &nitems);
		newnvp = nvpair_create_uint32_array(name, data, nitems);
		break;
	case NV_TYPE_INT64_ARRAY:
		data = nvpair_get_int64_array(nvp, &nitems);
		newnvp = nvpair_create_int64_array(name, data, nitems);
		break;
	case NV_TYPE_UINT64_ARRAY:
		data = nvpair_get_uint64_array(nvp, &nitems);
		newnvp = nvpair_create_uint64_array(name, data, nitems);
		break;
	case NV_TYPE_STRING_ARRAY:
		data = nvpair_get_string_array(nvp, &nitems);
		newnvp = nvpair_create_string_array(name, data, nitems);
		break;
	case NV_TYPE_NVLIST_ARRAY:
		data = nvpair_get_nvlist_array(nvp, &nitems);
		newnvp = nvpair_create_nvlist_array(name, data, nitems);
		break;
	case NV_TYPE_DESCRIPTOR_ARRAY:
		data = nvpair_get_descriptor_array(nvp, &nitems);
		newnvp = nvpair_create_descriptor_array(name, data, nitems);
		break;
	default:
		PJDLOG_ABORT("Unknown type: %d.", nvpair_type(nvp));
	}

	return (newnvp);
}

size_t
nvpair_header_size(void)
{

	return (sizeof(struct nvpair_header));
}

size_t
nvpair_size(const nvpair_t *nvp)
{

	NVPAIR_ASSERT(nvp);

	return (nvp->nvp_datasize);
}

static unsigned char *
nvpair_pack_header(const nvpair_t *nvp, unsigned char *ptr, size_t *leftp)
{
	struct nvpair_header nvphdr;
	size_t namesize;

	NVPAIR_ASSERT(nvp);

	nvphdr.nvph_type = nvp->nvp_type;
	namesize = strlen(nvp->nvp_name) + 1;
	PJDLOG_ASSERT(namesize > 0 && namesize <= UINT16_MAX);
	nvphdr.nvph_namesize = namesize;
	nvphdr.nvph_datasize = nvp->nvp_datasize;
	PJDLOG_ASSERT(*leftp >= sizeof(nvphdr));
	memcpy(ptr, &nvphdr, sizeof(nvphdr));
	ptr += sizeof(nvphdr);
	*leftp -= sizeof(nvphdr);

	PJDLOG_ASSERT(*leftp >= namesize);
	memcpy(ptr, nvp->nvp_name, namesize);
	ptr += namesize;
	*leftp -= namesize;

	return (ptr);
}

static unsigned char *
nvpair_pack_null(const nvpair_t *nvp, unsigned char *ptr,
    size_t *leftp __unused)
{

	NVPAIR_ASSERT(nvp);
	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_NULL);

	return (ptr);
}

static unsigned char *
nvpair_pack_bool(const nvpair_t *nvp, unsigned char *ptr, size_t *leftp)
{
	uint8_t value;

	NVPAIR_ASSERT(nvp);
	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_BOOL);

	value = (uint8_t)nvp->nvp_data;

	PJDLOG_ASSERT(*leftp >= sizeof(value));
	memcpy(ptr, &value, sizeof(value));
	ptr += sizeof(value);
	*leftp -= sizeof(value);

	return (ptr);
}

#define	NVPAIR_PACK(type, TYPE)						\
static unsigned char *							\
nvpair_pack_##type(const nvpair_t *nvp, unsigned char *ptr,		\
    size_t *leftp)							\
{									\
	type##_t value;							\
									\
	NVPAIR_ASSERT(nvp);						\
	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_##TYPE);			\
									\
	value = (type##_t)nvp->nvp_data;				\
									\
	PJDLOG_ASSERT(*leftp >= sizeof(value));				\
	memcpy(ptr, &value, sizeof(value));				\
	ptr += sizeof(value);						\
	*leftp -= sizeof(value);					\
									\
	return (ptr);							\
}

NVPAIR_PACK(int8, INT8)
NVPAIR_PACK(uint8, UINT8)
NVPAIR_PACK(int16, INT16)
NVPAIR_PACK(uint16, UINT16)
NVPAIR_PACK(int32, INT32)
NVPAIR_PACK(uint32, UINT32)
NVPAIR_PACK(int64, INT64)
NVPAIR_PACK(uint64, UINT64)

static unsigned char *
nvpair_pack_common_array(const nvpair_t *nvp, unsigned char *ptr, size_t *leftp)
{

	NVPAIR_ASSERT(nvp);

	if (nvp->nvp_datasize == 0)
		return (ptr);

	PJDLOG_ASSERT(*leftp >= nvp->nvp_datasize);
	memcpy(ptr, (const void *)(intptr_t)nvp->nvp_data, nvp->nvp_datasize);
	ptr += nvp->nvp_datasize;
	*leftp -= nvp->nvp_datasize;

	return (ptr);
}

static unsigned char *
nvpair_pack_string(const nvpair_t *nvp, unsigned char *ptr, size_t *leftp)
{

	NVPAIR_ASSERT(nvp);
	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_STRING);

	return (nvpair_pack_common_array(nvp, ptr, leftp));
}

static unsigned char *
nvpair_pack_nvlist(const nvpair_t *nvp, unsigned char *ptr, int64_t *fdidxp,
    size_t *leftp)
{
	unsigned char *data;
	size_t size;

	NVPAIR_ASSERT(nvp);
	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_NVLIST);

	if (nvp->nvp_datasize == 0)
		return (ptr);

	data = nvlist_xpack((const nvlist_t *)(intptr_t)nvp->nvp_data, fdidxp,
	    &size);
	if (data == NULL)
		return (NULL);

	PJDLOG_ASSERT(size == nvp->nvp_datasize);
	PJDLOG_ASSERT(*leftp >= nvp->nvp_datasize);

	memcpy(ptr, data, nvp->nvp_datasize);
	free(data);

	ptr += nvp->nvp_datasize;
	*leftp -= nvp->nvp_datasize;

	return (ptr);
}

static unsigned char *
nvpair_pack_descriptor(const nvpair_t *nvp, unsigned char *ptr, int64_t *fdidxp,
    size_t *leftp)
{
	int64_t value;

	NVPAIR_ASSERT(nvp);
	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_DESCRIPTOR);

	value = (int64_t)nvp->nvp_data;
	if (value != -1) {
		/*
		 * If there is a real descriptor here, we change its number
		 * to position in the array of descriptors send via control
		 * message.
		 */
		PJDLOG_ASSERT(fdidxp != NULL);

		value = *fdidxp;
		(*fdidxp)++;
	}

	PJDLOG_ASSERT(*leftp >= sizeof(value));
	memcpy(ptr, &value, sizeof(value));
	ptr += sizeof(value);
	*leftp -= sizeof(value);

	return (ptr);
}

static unsigned char *
nvpair_pack_bool_array(const nvpair_t *nvp, unsigned char *ptr, size_t *leftp)
{
	unsigned int ii;
	uint8_t *values;
	const bool *bools;

	NVPAIR_ASSERT(nvp);
	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_BOOL_ARRAY);

	if (nvp->nvp_datasize == 0)
		return (ptr);

	values = malloc(nvp->nvp_datasize);
	if (values == NULL) {
		free(ptr);
		return (NULL);
	}
	bools = (const bool *)(intptr_t)nvp->nvp_data;
	for (ii = 0; ii < nvp->nvp_nitems; ii++)
		values[ii] = bools[ii] ? 1 : 0;

	PJDLOG_ASSERT(*leftp >= nvp->nvp_datasize);
	memcpy(ptr, values, nvp->nvp_datasize);
	ptr += nvp->nvp_datasize;
	*leftp -= nvp->nvp_datasize;

	free(values);

	return (ptr);
}

static unsigned char *
nvpair_pack_string_array(const nvpair_t *nvp, unsigned char *ptr, size_t *leftp)
{
	const char * const *dataptr;
	const char *data;
	uint64_t nitems, strsize;
	size_t datasize;
	unsigned int ii;

	NVPAIR_ASSERT(nvp);
	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_STRING_ARRAY);

	if (nvp->nvp_datasize == 0)
		return (ptr);

	/*
	 * First we store number of items in the array.
	 */
	PJDLOG_ASSERT(*leftp >= sizeof(nitems));
	nitems = nvp->nvp_nitems;
	memcpy(ptr, &nitems, sizeof(nitems));
	ptr += sizeof(nitems);
	*leftp -= sizeof(nitems);

	dataptr = (const char * const *)(intptr_t)nvp->nvp_data;

	/*
	 * Then we store an array with string sizes (not length exactly:
	 * size of 0 is NULL element, size of 1 is an empty string).
	 */
	PJDLOG_ASSERT(*leftp >= sizeof(strsize) * nitems);
	for (ii = 0; ii < nitems; ii++) {
		if (dataptr[ii] == NULL)
			strsize = 0;
		else
			strsize = strlen(dataptr[ii]) + 1;
		memcpy(ptr, &strsize, sizeof(strsize));
		ptr += sizeof(strsize);
		*leftp -= sizeof(strsize);
	}

	/*
	 * And we copy the data.
	 */
	data = (const char *)(intptr_t)nvp->nvp_data + sizeof(strsize) * nitems;
	datasize = nvp->nvp_datasize - sizeof(nitems) -
	    sizeof(strsize) * nitems;
	PJDLOG_ASSERT(*leftp >= datasize);
	memcpy(ptr, data, datasize);
	ptr += datasize;
	*leftp -= datasize;

	return (ptr);
}

static unsigned char *
nvpair_pack_nvlist_array(const nvpair_t *nvp, unsigned char *ptr,
    int64_t *fdidxp __unused, size_t *leftp __unused)
{

	NVPAIR_ASSERT(nvp);
	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_NVLIST_ARRAY);

	if (nvp->nvp_datasize == 0)
		return (ptr);

	PJDLOG_ABORT("TODO");
	/* TODO */

	return (ptr);
}

static unsigned char *
nvpair_pack_descriptor_array(const nvpair_t *nvp, unsigned char *ptr,
    int64_t *fdidxp __unused, size_t *leftp __unused)
{

	NVPAIR_ASSERT(nvp);
	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_DESCRIPTOR_ARRAY);

	if (nvp->nvp_datasize == 0)
		return (ptr);

	PJDLOG_ABORT("TODO");
	/* TODO */

	return (ptr);
}

unsigned char *
nvpair_pack(nvpair_t *nvp, unsigned char *ptr, int64_t *fdidxp, size_t *leftp)
{
	const nvlist_t * const *nvls;
	unsigned int ii;

	NVPAIR_ASSERT(nvp);

	/*
	 * We have to update datasize for those two on every pack,
	 * so that proper datasize is placed into nvpair_header
	 * during the nvpair_pack_header() call below.
	 */
	switch (nvp->nvp_type) {
	case NV_TYPE_NVLIST:
		if (nvp->nvp_data == 0) {
			nvp->nvp_datasize = 0;
		} else {
			nvp->nvp_datasize =
			    nvlist_size((const nvlist_t *)(intptr_t)nvp->nvp_data);
		}
		break;
	case NV_TYPE_NVLIST_ARRAY:
		nvp->nvp_datasize = 0;
		nvls = (const nvlist_t * const *)(intptr_t)nvp->nvp_data;
		if (nvls == NULL)
			break;
		for (ii = 0; ii < nvp->nvp_nitems; ii++) {
			if (nvls[ii] == NULL)
				continue;
			nvp->nvp_datasize += nvlist_size(nvls[ii]);
		}
		break;
	}

	ptr = nvpair_pack_header(nvp, ptr, leftp);
	if (ptr == NULL)
		return (NULL);

	switch (nvp->nvp_type) {
	case NV_TYPE_NULL:
		ptr = nvpair_pack_null(nvp, ptr, leftp);
		break;
	case NV_TYPE_BOOL:
		ptr = nvpair_pack_bool(nvp, ptr, leftp);
		break;
	case NV_TYPE_INT8:
		ptr = nvpair_pack_int8(nvp, ptr, leftp);
		break;
	case NV_TYPE_UINT8:
		ptr = nvpair_pack_uint8(nvp, ptr, leftp);
		break;
	case NV_TYPE_INT16:
		ptr = nvpair_pack_int16(nvp, ptr, leftp);
		break;
	case NV_TYPE_UINT16:
		ptr = nvpair_pack_uint16(nvp, ptr, leftp);
		break;
	case NV_TYPE_INT32:
		ptr = nvpair_pack_int32(nvp, ptr, leftp);
		break;
	case NV_TYPE_UINT32:
		ptr = nvpair_pack_uint32(nvp, ptr, leftp);
		break;
	case NV_TYPE_INT64:
		ptr = nvpair_pack_int64(nvp, ptr, leftp);
		break;
	case NV_TYPE_UINT64:
		ptr = nvpair_pack_uint64(nvp, ptr, leftp);
		break;
	case NV_TYPE_STRING:
		ptr = nvpair_pack_string(nvp, ptr, leftp);
		break;
	case NV_TYPE_NVLIST:
		ptr = nvpair_pack_nvlist(nvp, ptr, fdidxp, leftp);
		break;
	case NV_TYPE_DESCRIPTOR:
		ptr = nvpair_pack_descriptor(nvp, ptr, fdidxp, leftp);
		break;
	case NV_TYPE_BOOL_ARRAY:
		ptr = nvpair_pack_bool_array(nvp, ptr, leftp);
		break;
	case NV_TYPE_INT8_ARRAY:
	case NV_TYPE_UINT8_ARRAY:
	case NV_TYPE_INT16_ARRAY:
	case NV_TYPE_UINT16_ARRAY:
	case NV_TYPE_INT32_ARRAY:
	case NV_TYPE_UINT32_ARRAY:
	case NV_TYPE_INT64_ARRAY:
	case NV_TYPE_UINT64_ARRAY:
		ptr = nvpair_pack_common_array(nvp, ptr, leftp);
		break;
	case NV_TYPE_STRING_ARRAY:
		ptr = nvpair_pack_string_array(nvp, ptr, leftp);
		break;
	case NV_TYPE_NVLIST_ARRAY:
		ptr = nvpair_pack_nvlist_array(nvp, ptr, fdidxp, leftp);
		break;
	case NV_TYPE_DESCRIPTOR_ARRAY:
		ptr = nvpair_pack_descriptor_array(nvp, ptr, fdidxp, leftp);
		break;
	default:
		PJDLOG_ABORT("Invalid type (%d).", nvp->nvp_type);
	}

	return (ptr);
}

static const unsigned char *
nvpair_unpack_header(int flags, nvpair_t *nvp, const unsigned char *ptr,
    size_t *leftp)
{
	struct nvpair_header nvphdr;

	if (*leftp < sizeof(nvphdr))
		goto failed;

	memcpy(&nvphdr, ptr, sizeof(nvphdr));
	ptr += sizeof(nvphdr);
	*leftp -= sizeof(nvphdr);

#if NV_TYPE_FIRST > 0
	if (nvphdr.nvph_type < NV_TYPE_FIRST)
		goto failed;
#endif
	if (nvphdr.nvph_type > NV_TYPE_LAST)
		goto failed;

#if BYTE_ORDER == BIG_ENDIAN
	if ((flags & NV_FLAG_BIG_ENDIAN) == 0) {
		nvphdr.nvph_namesize = le16toh(nvphdr.nvph_namesize);
		nvphdr.nvph_datasize = le64toh(nvphdr.nvph_datasize);
	}
#else
	if ((flags & NV_FLAG_BIG_ENDIAN) != 0) {
		nvphdr.nvph_namesize = be16toh(nvphdr.nvph_namesize);
		nvphdr.nvph_datasize = be64toh(nvphdr.nvph_datasize);
	}
#endif

	if (*leftp < nvphdr.nvph_namesize)
		goto failed;
	if (nvphdr.nvph_namesize < 1)
		goto failed;
	if (strnlen((const char *)ptr, nvphdr.nvph_namesize) !=
	    (size_t)(nvphdr.nvph_namesize - 1)) {
		goto failed;
	}

	memcpy(nvp->nvp_name, ptr, nvphdr.nvph_namesize);
	ptr += nvphdr.nvph_namesize;
	*leftp -= nvphdr.nvph_namesize;

	if (*leftp < nvphdr.nvph_datasize)
		goto failed;

	nvp->nvp_type = nvphdr.nvph_type;
	nvp->nvp_data = 0;
	nvp->nvp_datasize = nvphdr.nvph_datasize;
	nvp->nvp_nitems = 0;

	return (ptr);
failed:
	errno = EINVAL;
	return (NULL);
}

static const unsigned char *
nvpair_unpack_null(int flags __unused, nvpair_t *nvp, const unsigned char *ptr,
    size_t *leftp __unused)
{

	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_NULL);

	if (nvp->nvp_datasize != 0) {
		errno = EINVAL;
		return (NULL);
	}

	return (ptr);
}

static const unsigned char *
nvpair_unpack_bool(int flags __unused, nvpair_t *nvp, const unsigned char *ptr,
    size_t *leftp)
{
	uint8_t value;

	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_BOOL);

	if (nvp->nvp_datasize != sizeof(value)) {
		errno = EINVAL;
		return (NULL);
	}
	if (*leftp < sizeof(value)) {
		errno = EINVAL;
		return (NULL);
	}

	memcpy(&value, ptr, sizeof(value));
	ptr += sizeof(value);
	*leftp -= sizeof(value);

	if (value != 0 && value != 1) {
		errno = EINVAL;
		return (NULL);
	}

	nvp->nvp_data = (uint64_t)value;

	return (ptr);
}

static const unsigned char *
nvpair_unpack_int8(int flags __unused, nvpair_t *nvp, const unsigned char *ptr,
    size_t *leftp)
{
	int8_t value;

	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_INT8);

	if (nvp->nvp_datasize != sizeof(value)) {
		errno = EINVAL;
		return (NULL);
	}
	if (*leftp < sizeof(value)) {
		errno = EINVAL;
		return (NULL);
	}

	memcpy(&value, ptr, sizeof(value));
	ptr += sizeof(value);
	*leftp -= sizeof(value);

	nvp->nvp_data = (uint64_t)value;

	return (ptr);
}

static const unsigned char *
nvpair_unpack_uint8(int flags __unused, nvpair_t *nvp, const unsigned char *ptr,
    size_t *leftp)
{
	uint8_t value;

	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_UINT8);

	if (nvp->nvp_datasize != sizeof(value)) {
		errno = EINVAL;
		return (NULL);
	}
	if (*leftp < sizeof(value)) {
		errno = EINVAL;
		return (NULL);
	}

	memcpy(&value, ptr, sizeof(value));
	ptr += sizeof(value);
	*leftp -= sizeof(value);

	nvp->nvp_data = (uint64_t)value;

	return (ptr);
}

static const unsigned char *
nvpair_unpack_int16(int flags, nvpair_t *nvp, const unsigned char *ptr,
    size_t *leftp)
{

	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_INT16);

	if (nvp->nvp_datasize != sizeof(int16_t)) {
		errno = EINVAL;
		return (NULL);
	}
	if (*leftp < sizeof(int16_t)) {
		errno = EINVAL;
		return (NULL);
	}

	if ((flags & NV_FLAG_BIG_ENDIAN) != 0)
		nvp->nvp_data = (uint64_t)be16dec(ptr);
	else
		nvp->nvp_data = (uint64_t)le16dec(ptr);
	ptr += sizeof(int16_t);
	*leftp -= sizeof(int16_t);

	return (ptr);
}

static const unsigned char *
nvpair_unpack_uint16(int flags, nvpair_t *nvp, const unsigned char *ptr,
    size_t *leftp)
{

	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_UINT16);

	if (nvp->nvp_datasize != sizeof(uint16_t)) {
		errno = EINVAL;
		return (NULL);
	}
	if (*leftp < sizeof(uint16_t)) {
		errno = EINVAL;
		return (NULL);
	}

	if ((flags & NV_FLAG_BIG_ENDIAN) != 0)
		nvp->nvp_data = (uint64_t)be16dec(ptr);
	else
		nvp->nvp_data = (uint64_t)le16dec(ptr);
	ptr += sizeof(uint16_t);
	*leftp -= sizeof(uint16_t);

	return (ptr);
}

static const unsigned char *
nvpair_unpack_int32(int flags, nvpair_t *nvp, const unsigned char *ptr,
    size_t *leftp)
{

	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_INT32);

	if (nvp->nvp_datasize != sizeof(int32_t)) {
		errno = EINVAL;
		return (NULL);
	}
	if (*leftp < sizeof(int32_t)) {
		errno = EINVAL;
		return (NULL);
	}

	if ((flags & NV_FLAG_BIG_ENDIAN) != 0)
		nvp->nvp_data = (uint64_t)be32dec(ptr);
	else
		nvp->nvp_data = (uint64_t)le32dec(ptr);
	ptr += sizeof(int32_t);
	*leftp -= sizeof(int32_t);

	return (ptr);
}

static const unsigned char *
nvpair_unpack_uint32(int flags, nvpair_t *nvp, const unsigned char *ptr,
    size_t *leftp)
{

	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_UINT32);

	if (nvp->nvp_datasize != sizeof(uint32_t)) {
		errno = EINVAL;
		return (NULL);
	}
	if (*leftp < sizeof(uint32_t)) {
		errno = EINVAL;
		return (NULL);
	}

	if ((flags & NV_FLAG_BIG_ENDIAN) != 0)
		nvp->nvp_data = (uint64_t)be32dec(ptr);
	else
		nvp->nvp_data = (uint64_t)le32dec(ptr);
	ptr += sizeof(uint32_t);
	*leftp -= sizeof(uint32_t);

	return (ptr);
}

static const unsigned char *
nvpair_unpack_int64(int flags, nvpair_t *nvp, const unsigned char *ptr,
    size_t *leftp)
{

	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_INT64);

	if (nvp->nvp_datasize != sizeof(int64_t)) {
		errno = EINVAL;
		return (NULL);
	}
	if (*leftp < sizeof(int64_t)) {
		errno = EINVAL;
		return (NULL);
	}

	if ((flags & NV_FLAG_BIG_ENDIAN) != 0)
		nvp->nvp_data = be64dec(ptr);
	else
		nvp->nvp_data = le64dec(ptr);
	ptr += sizeof(int64_t);
	*leftp -= sizeof(int64_t);

	return (ptr);
}

static const unsigned char *
nvpair_unpack_uint64(int flags, nvpair_t *nvp, const unsigned char *ptr,
    size_t *leftp)
{

	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_UINT64);

	if (nvp->nvp_datasize != sizeof(uint64_t)) {
		errno = EINVAL;
		return (NULL);
	}
	if (*leftp < sizeof(uint64_t)) {
		errno = EINVAL;
		return (NULL);
	}

	if ((flags & NV_FLAG_BIG_ENDIAN) != 0)
		nvp->nvp_data = be64dec(ptr);
	else
		nvp->nvp_data = le64dec(ptr);
	ptr += sizeof(uint64_t);
	*leftp -= sizeof(uint64_t);

	return (ptr);
}

static const unsigned char *
nvpair_unpack_string(int flags, nvpair_t *nvp, const unsigned char *ptr,
    size_t *leftp)
{

	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_STRING);

	if (*leftp < nvp->nvp_datasize) {
		errno = EINVAL;
		return (NULL);
	}
	if (nvp->nvp_datasize == 0) {
		if ((flags & NV_FLAG_NO_VALUE) == 0) {
			errno = EINVAL;
			return (NULL);
		}
		nvp->nvp_data = 0;
	} else {
		if (strnlen((const char *)ptr, nvp->nvp_datasize) !=
		    nvp->nvp_datasize - 1) {
			errno = EINVAL;
			return (NULL);
		}

		nvp->nvp_data = (uint64_t)(uintptr_t)strdup((const char *)ptr);
		if (nvp->nvp_data == 0)
			return (NULL);
	}

	ptr += nvp->nvp_datasize;
	*leftp -= nvp->nvp_datasize;

	return (ptr);
}

static const unsigned char *
nvpair_unpack_nvlist(int flags, nvpair_t *nvp, const unsigned char *ptr,
    size_t *leftp, const int *fds, size_t nfds)
{
	nvlist_t *value;

	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_NVLIST);

	if (*leftp < nvp->nvp_datasize) {
		errno = EINVAL;
		return (NULL);
	}

	if (nvp->nvp_datasize == 0) {
		if ((flags & NV_FLAG_NO_VALUE) == 0) {
			errno = EINVAL;
			return (NULL);
		}
		return (ptr);
	}

	value = nvlist_xunpack(ptr, nvp->nvp_datasize, fds, nfds);
	if (value == NULL)
		return (NULL);

	nvp->nvp_data = (uint64_t)(uintptr_t)value;
	nvp->nvp_nitems = 0;

	ptr += nvp->nvp_datasize;
	*leftp -= nvp->nvp_datasize;

	return (ptr);
}

static const unsigned char *
nvpair_unpack_descriptor(int flags, nvpair_t *nvp, const unsigned char *ptr,
    size_t *leftp, const int *fds, size_t nfds)
{
	int64_t idx;

	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_DESCRIPTOR);

	if (nvp->nvp_datasize != sizeof(idx)) {
		errno = EINVAL;
		return (NULL);
	}
	if (*leftp < sizeof(idx)) {
		errno = EINVAL;
		return (NULL);
	}

	if ((flags & NV_FLAG_BIG_ENDIAN) != 0)
		idx = be64dec(ptr);
	else
		idx = le64dec(ptr);

	if (idx == -1) {
		if ((flags & NV_FLAG_NO_VALUE) == 0) {
			errno = EINVAL;
			return (NULL);
		}
		nvp->nvp_data = (uint64_t)-1;
	} else if (idx >= 0) {
		if ((size_t)idx >= nfds) {
			errno = EINVAL;
			return (NULL);
		}
		nvp->nvp_data = (uint64_t)fds[idx];
	} else {
		errno = EINVAL;
		return (NULL);
	}

	ptr += sizeof(idx);
	*leftp -= sizeof(idx);

	return (ptr);
}

static const unsigned char *
nvpair_unpack_bool_array(int flags, nvpair_t *nvp, const unsigned char *ptr,
    size_t *leftp)
{
	unsigned int ii;
	const uint8_t *data;
	bool *value;

	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_BOOL_ARRAY);

	if (*leftp < nvp->nvp_datasize) {
		errno = EINVAL;
		return (NULL);
	}

	if (nvp->nvp_datasize == 0) {
		if ((flags & NV_FLAG_NO_VALUE) == 0) {
			errno = EINVAL;
			return (NULL);
		}
		return (ptr);
	}

	value = malloc(sizeof(bool) * nvp->nvp_datasize);
	if (value == NULL)
		return (NULL);

	data = (const uint8_t *)ptr;

	for (ii = 0; ii < nvp->nvp_datasize; ii++) {
		if (data[ii] != 0 && data[ii] != 1) {
			free(value);
			errno = EINVAL;
			return (NULL);
		}
		value[ii] = data[ii] == 1 ? true : false;
	}

	nvp->nvp_data = (uint64_t)(uintptr_t)value;
	nvp->nvp_nitems = nvp->nvp_datasize;

	ptr += nvp->nvp_datasize;
	*leftp -= nvp->nvp_datasize;

	return (ptr);
}

static const unsigned char *
nvpair_unpack_common_array(int flags, size_t bytes, nvpair_t *nvp,
    const unsigned char *ptr, size_t *leftp)
{
	unsigned int ii;
	uint16_t *value;

	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_INT8_ARRAY ||
	    nvp->nvp_type == NV_TYPE_UINT8_ARRAY ||
	    nvp->nvp_type == NV_TYPE_INT16_ARRAY ||
	    nvp->nvp_type == NV_TYPE_UINT16_ARRAY ||
	    nvp->nvp_type == NV_TYPE_INT32_ARRAY ||
	    nvp->nvp_type == NV_TYPE_UINT32_ARRAY ||
	    nvp->nvp_type == NV_TYPE_INT64_ARRAY ||
	    nvp->nvp_type == NV_TYPE_UINT64_ARRAY);

	if (*leftp < nvp->nvp_datasize) {
		errno = EINVAL;
		return (NULL);
	}
	if ((nvp->nvp_datasize % bytes) != 0) {
		errno = EINVAL;
		return (NULL);
	}

	if (nvp->nvp_datasize == 0) {
		if ((flags & NV_FLAG_NO_VALUE) == 0) {
			errno = EINVAL;
			return (NULL);
		}
		return (ptr);
	}

	value = malloc(sizeof(nvp->nvp_datasize));
	if (value == NULL)
		return (NULL);

	nvp->nvp_nitems = nvp->nvp_datasize / bytes;

#if BYTE_ORDER == BIG_ENDIAN
	if (bytes == 1 || (flags & NV_FLAG_BIG_ENDIAN) != 0) {
		memcpy(value, ptr, nvp->nvp_datasize);
		ptr += nvp->nvp_datasize;
		*leftp -= nvp->nvp_datasize;
	} else {
		for (ii = 0; ii < nvp->nvp_nitems; ii++) {
			switch (bytes) {
			case 2:
				value[ii] = le16dec(ptr);
				break;
			case 4:
				value[ii] = le32dec(ptr);
				break;
			case 8:
				value[ii] = le64dec(ptr);
				break;
			default:
				PJDLOG_ABORT("Unexpected size (%zu).", bytes);
			}
			ptr += sizeof(value[ii]);
			*leftp -= sizeof(value[ii]);
		}
	}
#else
	if (bytes == 1 || (flags & NV_FLAG_BIG_ENDIAN) == 0) {
		memcpy(value, ptr, nvp->nvp_datasize);
		ptr += nvp->nvp_datasize;
		*leftp -= nvp->nvp_datasize;
	} else {
		for (ii = 0; ii < nvp->nvp_nitems; ii++) {
			switch (bytes) {
			case 2:
				value[ii] = be16dec(ptr);
				break;
			case 4:
				value[ii] = be32dec(ptr);
				break;
			case 8:
				value[ii] = be64dec(ptr);
				break;
			default:
				PJDLOG_ABORT("Unexpected size (%zu).", bytes);
			}
			ptr += sizeof(value[ii]);
			*leftp -= sizeof(value[ii]);
		}
	}
#endif

	nvp->nvp_data = (uint64_t)(uintptr_t)value;

	return (ptr);
}

static const unsigned char *
nvpair_unpack_int8_array(int flags, nvpair_t *nvp, const unsigned char *ptr,
    size_t *leftp)
{

	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_INT8_ARRAY);

	return (nvpair_unpack_common_array(flags, sizeof(int8_t), nvp, ptr,
	    leftp));
}

static const unsigned char *
nvpair_unpack_uint8_array(int flags, nvpair_t *nvp, const unsigned char *ptr,
    size_t *leftp)
{

	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_UINT8_ARRAY);

	return (nvpair_unpack_common_array(flags, sizeof(uint8_t), nvp, ptr,
	    leftp));
}

static const unsigned char *
nvpair_unpack_int16_array(int flags, nvpair_t *nvp, const unsigned char *ptr,
    size_t *leftp)
{

	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_INT16_ARRAY);

	return (nvpair_unpack_common_array(flags, sizeof(int16_t), nvp, ptr,
	    leftp));
}

static const unsigned char *
nvpair_unpack_uint16_array(int flags, nvpair_t *nvp, const unsigned char *ptr,
    size_t *leftp)
{

	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_UINT16_ARRAY);

	return (nvpair_unpack_common_array(flags, sizeof(uint16_t), nvp, ptr,
	    leftp));
}

static const unsigned char *
nvpair_unpack_int32_array(int flags, nvpair_t *nvp, const unsigned char *ptr,
    size_t *leftp)
{

	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_INT32_ARRAY);

	return (nvpair_unpack_common_array(flags, sizeof(int32_t), nvp, ptr,
	    leftp));
}

static const unsigned char *
nvpair_unpack_uint32_array(int flags, nvpair_t *nvp, const unsigned char *ptr,
    size_t *leftp)
{

	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_UINT32_ARRAY);

	return (nvpair_unpack_common_array(flags, sizeof(uint32_t), nvp, ptr,
	    leftp));
}

static const unsigned char *
nvpair_unpack_int64_array(int flags, nvpair_t *nvp, const unsigned char *ptr,
    size_t *leftp)
{

	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_INT64_ARRAY);

	return (nvpair_unpack_common_array(flags, sizeof(int64_t), nvp, ptr,
	    leftp));
}

static const unsigned char *
nvpair_unpack_uint64_array(int flags, nvpair_t *nvp, const unsigned char *ptr,
    size_t *leftp)
{

	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_UINT64_ARRAY);

	return (nvpair_unpack_common_array(flags, sizeof(uint64_t), nvp, ptr,
	    leftp));
}

static const unsigned char *
nvpair_unpack_string_array(int flags, nvpair_t *nvp, const unsigned char *ptr,
    size_t *leftp)
{
	const unsigned char *sizes;
	uint64_t nitems, strsize;
	char **strs, *str;
	unsigned int ii;

	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_STRING_ARRAY);

	if (*leftp < nvp->nvp_datasize) {
		errno = EINVAL;
		return (NULL);
	}

	if (nvp->nvp_datasize == 0) {
		if ((flags & NV_FLAG_NO_VALUE) == 0) {
			errno = EINVAL;
			return (NULL);
		}
		return (ptr);
	}

	if (nvp->nvp_datasize < sizeof(nitems)) {
		errno = EINVAL;
		return (NULL);
	}

	if ((flags & NV_FLAG_BIG_ENDIAN) != 0)
		nitems = be64dec(ptr);
	else
		nitems = le64dec(ptr);
	ptr += sizeof(nitems);
	*leftp -= sizeof(nitems);

	if (nvp->nvp_datasize < sizeof(strsize) * nitems) {
		errno = EINVAL;
		return (NULL);
	}

	sizes = ptr;
	ptr += sizeof(strsize) * nitems;
	*leftp -= sizeof(strsize) * nitems;

	strs = malloc(sizeof(strs[0]) * nitems + nvp->nvp_datasize -
	    sizeof(nitems) - sizeof(strsize) * nitems);
	if (strs == NULL)
		return (NULL);
	str = (char *)(strs + nitems);

	for (ii = 0; ii < nitems; ii++) {
		if ((flags & NV_FLAG_BIG_ENDIAN) != 0)
			strsize = be64dec(sizes);
		else
			strsize = le64dec(sizes);
		sizes += sizeof(strsize);

		if (strsize == 0) {
			strs[ii] = NULL;
		} else {
			if (*leftp < strsize) {
				free(strs);
				errno = EINVAL;
				return (NULL);
			}
			if (strnlen((const char *)ptr, strsize) != strsize - 1) {
				free(strs);
				errno = EINVAL;
				return (NULL);
			}
			memcpy(str, (const char *)ptr, strsize);
			strs[ii] = str;
			str += strsize;
			ptr += strsize;
			*leftp -= strsize;
		}
	}

	nvp->nvp_data = (uint64_t)(uintptr_t)strs;
	nvp->nvp_nitems = nitems;

	return (ptr);
}

static const unsigned char *
nvpair_unpack_nvlist_array(int flags __unused, nvpair_t *nvp,
    const unsigned char *ptr, size_t *leftp __unused, const int *fds __unused,
    size_t nfds __unused)
{

	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_NVLIST_ARRAY);

	PJDLOG_ABORT("TODO");
	/* TODO */

	return (ptr);
}

static const unsigned char *
nvpair_unpack_descriptor_array(int flags __unused, nvpair_t *nvp,
    const unsigned char *ptr, size_t *leftp __unused, const int *fds __unused,
    size_t nfds __unused)
{

	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_DESCRIPTOR_ARRAY);

	PJDLOG_ABORT("TODO");
	/* TODO */

	return (ptr);
}

const unsigned char *
nvpair_unpack(int flags, const unsigned char *ptr, size_t *leftp,
    const int *fds, size_t nfds, nvpair_t **nvpp)
{
	nvpair_t *nvp, *tmp;

	nvp = calloc(1, sizeof(*nvp) + NVPAIR_NAME_MAX);
	if (nvp == NULL)
		return (NULL);
	nvp->nvp_name = (char *)(nvp + 1);

	ptr = nvpair_unpack_header(flags, nvp, ptr, leftp);
	if (ptr == NULL)
		goto failed;
	tmp = realloc(nvp, sizeof(*nvp) + strlen(nvp->nvp_name) + 1);
	if (tmp == NULL) {
		free(nvp);
		goto failed;
	}
	nvp = tmp;
	/* Update nvp_name after realloc(). */
	nvp->nvp_name = (char *)(nvp + 1);

	switch (nvp->nvp_type) {
	case NV_TYPE_NULL:
		ptr = nvpair_unpack_null(flags, nvp, ptr, leftp);
		break;
	case NV_TYPE_BOOL:
		ptr = nvpair_unpack_bool(flags, nvp, ptr, leftp);
		break;
	case NV_TYPE_INT8:
		ptr = nvpair_unpack_int8(flags, nvp, ptr, leftp);
		break;
	case NV_TYPE_UINT8:
		ptr = nvpair_unpack_uint8(flags, nvp, ptr, leftp);
		break;
	case NV_TYPE_INT16:
		ptr = nvpair_unpack_int16(flags, nvp, ptr, leftp);
		break;
	case NV_TYPE_UINT16:
		ptr = nvpair_unpack_uint16(flags, nvp, ptr, leftp);
		break;
	case NV_TYPE_INT32:
		ptr = nvpair_unpack_int32(flags, nvp, ptr, leftp);
		break;
	case NV_TYPE_UINT32:
		ptr = nvpair_unpack_uint32(flags, nvp, ptr, leftp);
		break;
	case NV_TYPE_INT64:
		ptr = nvpair_unpack_int64(flags, nvp, ptr, leftp);
		break;
	case NV_TYPE_UINT64:
		ptr = nvpair_unpack_uint64(flags, nvp, ptr, leftp);
		break;
	case NV_TYPE_STRING:
		ptr = nvpair_unpack_string(flags, nvp, ptr, leftp);
		break;
	case NV_TYPE_NVLIST:
		ptr = nvpair_unpack_nvlist(flags, nvp, ptr, leftp, fds,
		    nfds);
		break;
	case NV_TYPE_DESCRIPTOR:
		ptr = nvpair_unpack_descriptor(flags, nvp, ptr, leftp, fds,
		    nfds);
		break;
	case NV_TYPE_BOOL_ARRAY:
		ptr = nvpair_unpack_bool_array(flags, nvp, ptr, leftp);
		break;
	case NV_TYPE_INT8_ARRAY:
		ptr = nvpair_unpack_int8_array(flags, nvp, ptr, leftp);
		break;
	case NV_TYPE_UINT8_ARRAY:
		ptr = nvpair_unpack_uint8_array(flags, nvp, ptr, leftp);
		break;
	case NV_TYPE_INT16_ARRAY:
		ptr = nvpair_unpack_int16_array(flags, nvp, ptr, leftp);
		break;
	case NV_TYPE_UINT16_ARRAY:
		ptr = nvpair_unpack_uint16_array(flags, nvp, ptr, leftp);
		break;
	case NV_TYPE_INT32_ARRAY:
		ptr = nvpair_unpack_int32_array(flags, nvp, ptr, leftp);
		break;
	case NV_TYPE_UINT32_ARRAY:
		ptr = nvpair_unpack_uint32_array(flags, nvp, ptr, leftp);
		break;
	case NV_TYPE_INT64_ARRAY:
		ptr = nvpair_unpack_int64_array(flags, nvp, ptr, leftp);
		break;
	case NV_TYPE_UINT64_ARRAY:
		ptr = nvpair_unpack_uint64_array(flags, nvp, ptr, leftp);
		break;
	case NV_TYPE_STRING_ARRAY:
		ptr = nvpair_unpack_string_array(flags, nvp, ptr, leftp);
		break;
	case NV_TYPE_NVLIST_ARRAY:
		ptr = nvpair_unpack_nvlist_array(flags, nvp, ptr, leftp,
		    fds, nfds);
		break;
	case NV_TYPE_DESCRIPTOR_ARRAY:
		ptr = nvpair_unpack_descriptor_array(flags, nvp, ptr,
		    leftp, fds, nfds);
		break;
	default:
		PJDLOG_ABORT("Invalid type (%d).", nvp->nvp_type);
	}

	if (ptr == NULL)
		goto failed;

	nvp->nvp_magic = NVPAIR_MAGIC;
	*nvpp = nvp;
	return (ptr);
failed:
	free(nvp);
	return (NULL);
}

int
nvpair_type(const nvpair_t *nvp)
{

	NVPAIR_ASSERT(nvp);

	return (nvp->nvp_type);
}

const char *
nvpair_name(const nvpair_t *nvp)
{

	NVPAIR_ASSERT(nvp);

	return (nvp->nvp_name);
}

static nvpair_t *
nvpair_allocv(int type, uint64_t data, size_t datasize, size_t nitems,
    const char *namefmt, va_list nameap)
{
	nvpair_t *nvp;
	char *name;
	int namelen;

	PJDLOG_ASSERT(type >= NV_TYPE_FIRST && type <= NV_TYPE_LAST);

	namelen = vasprintf(&name, namefmt, nameap);
	if (namelen < 0)
		return (NULL);

	PJDLOG_ASSERT(namelen > 0);
	if (namelen >= NVPAIR_NAME_MAX) {
		free(name);
		errno = ENAMETOOLONG;
		return (NULL);
	}

	nvp = calloc(1, sizeof(*nvp) + namelen + 1);
	if (nvp != NULL) {
		nvp->nvp_name = (char *)(nvp + 1);
		memcpy(nvp->nvp_name, name, namelen + 1);
		nvp->nvp_type = type;
		nvp->nvp_data = data;
		nvp->nvp_datasize = datasize;
		nvp->nvp_nitems = nitems;
		nvp->nvp_magic = NVPAIR_MAGIC;
	}
	free(name);

	return (nvp);
};

nvpair_t *
nvpair_create_null(const char *name)
{

	return (nvpair_createf_null("%s", name));
}

#define	NVPAIR_CREATE(vtype, type)					\
nvpair_t *								\
nvpair_create_##type(const char *name, vtype value)			\
{									\
									\
	return (nvpair_createf_##type(value, "%s", name));		\
}

NVPAIR_CREATE(bool, bool)
NVPAIR_CREATE(int8_t, int8)
NVPAIR_CREATE(uint8_t, uint8)
NVPAIR_CREATE(int16_t, int16)
NVPAIR_CREATE(uint16_t, uint16)
NVPAIR_CREATE(int32_t, int32)
NVPAIR_CREATE(uint32_t, uint32)
NVPAIR_CREATE(int64_t, int64)
NVPAIR_CREATE(uint64_t, uint64)
NVPAIR_CREATE(const char *, string)

nvpair_t *
nvpair_create_stringf(const char *name, const char *valuefmt, ...)
{
	va_list valueap;
	nvpair_t *nvp;

	va_start(valueap, valuefmt);
	nvp = nvpair_create_stringv(name, valuefmt, valueap);
	va_end(valueap);

	return (nvp);
}

nvpair_t *
nvpair_create_stringv(const char *name, const char *valuefmt, va_list valueap)
{
	nvpair_t *nvp;
	char *str;
	int len;

	len = vasprintf(&str, valuefmt, valueap);
	if (len < 0)
		return (NULL);
	nvp = nvpair_create_string(name, str);
	if (nvp == NULL)
		free(str);
	return (nvp);
}

NVPAIR_CREATE(const nvlist_t *, nvlist)
NVPAIR_CREATE(int, descriptor)

#define	NVPAIR_CREATE_ARRAY(vtype, type)				\
nvpair_t *								\
nvpair_create_##type##_array(const char *name, vtype value,		\
    size_t nitems)							\
{									\
									\
	return (nvpair_createf_##type##_array(value, nitems, "%s",	\
	    name));							\
}

NVPAIR_CREATE_ARRAY(const bool *, bool)
NVPAIR_CREATE_ARRAY(const int8_t *, int8)
NVPAIR_CREATE_ARRAY(const uint8_t *, uint8)
NVPAIR_CREATE_ARRAY(const int16_t *, int16)
NVPAIR_CREATE_ARRAY(const uint16_t *, uint16)
NVPAIR_CREATE_ARRAY(const int32_t *, int32)
NVPAIR_CREATE_ARRAY(const uint32_t *, uint32)
NVPAIR_CREATE_ARRAY(const int64_t *, int64)
NVPAIR_CREATE_ARRAY(const uint64_t *, uint64)
NVPAIR_CREATE_ARRAY(const char * const *, string)
NVPAIR_CREATE_ARRAY(const nvlist_t * const *, nvlist)
NVPAIR_CREATE_ARRAY(const int *, descriptor)

nvpair_t *
nvpair_createf_null(const char *namefmt, ...)
{
	va_list nameap;
	nvpair_t *nvp;

	va_start(nameap, namefmt);
	nvp = nvpair_createv_null(namefmt, nameap);
	va_end(nameap);

	return (nvp);
}

#define	NVPAIR_CREATEF(vtype, type)					\
nvpair_t *								\
nvpair_createf_##type(vtype value, const char *namefmt, ...)		\
{									\
	va_list nameap;							\
	nvpair_t *nvp;							\
									\
	va_start(nameap, namefmt);					\
	nvp = nvpair_createv_##type(value, namefmt, nameap);		\
	va_end(nameap);							\
									\
	return (nvp);							\
}

NVPAIR_CREATEF(bool, bool)
NVPAIR_CREATEF(int8_t, int8)
NVPAIR_CREATEF(uint8_t, uint8)
NVPAIR_CREATEF(int16_t, int16)
NVPAIR_CREATEF(uint16_t, uint16)
NVPAIR_CREATEF(int32_t, int32)
NVPAIR_CREATEF(uint32_t, uint32)
NVPAIR_CREATEF(int64_t, int64)
NVPAIR_CREATEF(uint64_t, uint64)
NVPAIR_CREATEF(const char *, string)
NVPAIR_CREATEF(const nvlist_t *, nvlist)
NVPAIR_CREATEF(int, descriptor)

#define	NVPAIR_CREATEF_ARRAY(vtype, type)				\
nvpair_t *								\
nvpair_createf_##type##_array(vtype value, size_t nitems,		\
    const char *namefmt, ...)						\
{									\
	va_list nameap;							\
	nvpair_t *nvp;							\
									\
	va_start(nameap, namefmt);					\
	nvp = nvpair_createv_##type##_array(value, nitems, namefmt,	\
	    nameap);							\
	va_end(nameap);							\
									\
	return (nvp);							\
}

NVPAIR_CREATEF_ARRAY(const bool *, bool)
NVPAIR_CREATEF_ARRAY(const int8_t *, int8)
NVPAIR_CREATEF_ARRAY(const uint8_t *, uint8)
NVPAIR_CREATEF_ARRAY(const int16_t *, int16)
NVPAIR_CREATEF_ARRAY(const uint16_t *, uint16)
NVPAIR_CREATEF_ARRAY(const int32_t *, int32)
NVPAIR_CREATEF_ARRAY(const uint32_t *, uint32)
NVPAIR_CREATEF_ARRAY(const int64_t *, int64)
NVPAIR_CREATEF_ARRAY(const uint64_t *, uint64)
NVPAIR_CREATEF_ARRAY(const char * const *, string)
NVPAIR_CREATEF_ARRAY(const nvlist_t * const *, nvlist)
NVPAIR_CREATEF_ARRAY(const int *, descriptor)

nvpair_t *
nvpair_createv_null(const char *namefmt, va_list nameap)
{

	return (nvpair_allocv(NV_TYPE_NULL, 0, 0, 0, namefmt, nameap));
}

nvpair_t *
nvpair_createv_bool(bool value, const char *namefmt, va_list nameap)
{

	return (nvpair_allocv(NV_TYPE_BOOL, value ? 1 : 0, sizeof(uint8_t), 0,
	    namefmt, nameap));
}

#define	NVPAIR_CREATEV(type, TYPE)					\
nvpair_t *								\
nvpair_createv_##type(type##_t value, const char *namefmt,		\
    va_list nameap)							\
{									\
									\
	return (nvpair_allocv(NV_TYPE_##TYPE, (uint64_t)value,		\
	    sizeof(value), 0, namefmt, nameap));			\
}

NVPAIR_CREATEV(int8, INT8)
NVPAIR_CREATEV(uint8, UINT8)
NVPAIR_CREATEV(int16, INT16)
NVPAIR_CREATEV(uint16, UINT16)
NVPAIR_CREATEV(int32, INT32)
NVPAIR_CREATEV(uint32, UINT32)
NVPAIR_CREATEV(int64, INT64)
NVPAIR_CREATEV(uint64, UINT64)

nvpair_t *
nvpair_createv_string(const char *value, const char *namefmt, va_list nameap)
{
	nvpair_t *nvp;
	size_t size;
	char *data;

	if (value == NULL) {
		data = NULL;
		size = 0;
	} else {
		data = strdup(value);
		if (data == NULL)
			return (NULL);
		size = strlen(value) + 1;
	}

	nvp = nvpair_allocv(NV_TYPE_STRING, (uint64_t)(uintptr_t)data, size, 0,
	    namefmt, nameap);
	if (nvp == NULL)
		free(data);

	return (nvp);
}

nvpair_t *
nvpair_createv_nvlist(const nvlist_t *value, const char *namefmt,
    va_list nameap)
{
	nvlist_t *nvl;
	nvpair_t *nvp;

	if (value == NULL) {
		nvl = NULL;
	} else {
		nvl = nvlist_clone(value);
		if (nvl == NULL)
			return (NULL);
	}
	nvp = nvpair_allocv(NV_TYPE_NVLIST, (uint64_t)(uintptr_t)nvl, 0, 0,
	    namefmt, nameap);
	if (nvp == NULL)
		nvlist_destroy(nvl);
	return (nvp);
}

nvpair_t *
nvpair_createv_descriptor(int value, const char *namefmt, va_list nameap)
{
	nvpair_t *nvp;

	PJDLOG_ASSERT(value == -1 || fd_is_valid(value));

	if (value != -1) {
		value = dup(value);
		if (value < 0)
			return (NULL);
	}
	nvp = nvpair_allocv(NV_TYPE_DESCRIPTOR, (uint64_t)value,
	    sizeof(int64_t), 0, namefmt, nameap);
	if (nvp == NULL && value != -1)
		close(value);
	return (nvp);
}

nvpair_t *
nvpair_createv_bool_array(const bool *value, size_t nitems, const char *namefmt,
    va_list nameap)
{
	nvpair_t *nvp;
	size_t size;
	void *data;

	PJDLOG_ASSERT((value == NULL && nitems == 0) ||
	    (value != NULL && nitems > 0));

	if (value == NULL) {
		size = 0;
		data = NULL;
	} else {
		size = sizeof(value[0]) * nitems;
		data = malloc(size);
		if (data == NULL)
			return (NULL);
		memcpy(data, value, size);
		size = sizeof(uint8_t) * nitems;
	}

	nvp = nvpair_allocv(NV_TYPE_BOOL_ARRAY, (uint64_t)(uintptr_t)data, size,
	    nitems, namefmt, nameap);
	if (nvp == NULL)
		free(data);

	return (nvp);
}

#define	NVPAIR_CREATEV_ARRAY(vtype, type, TYPE)				\
nvpair_t *								\
nvpair_createv_##type##_array(vtype value, size_t nitems,		\
    const char *namefmt, va_list nameap)				\
{									\
	nvpair_t *nvp;							\
	size_t size;							\
	void *data;							\
									\
	PJDLOG_ASSERT((value == NULL && nitems == 0) ||			\
	    (value != NULL && nitems > 0));				\
									\
	if (value == NULL) {						\
		size = 0;						\
		data = NULL;						\
	} else {							\
		size = sizeof(value[0]) * nitems;			\
		data = malloc(size);					\
		if (data == NULL)					\
			return (NULL);					\
		memcpy(data, value, size);				\
	}								\
									\
	nvp = nvpair_allocv(NV_TYPE_##TYPE##_ARRAY,			\
	    (uint64_t)(uintptr_t)data, size, nitems, namefmt, nameap);	\
	if (nvp == NULL)						\
		free(data);						\
									\
	return (nvp);							\
}

NVPAIR_CREATEV_ARRAY(const int8_t *, int8, INT8)
NVPAIR_CREATEV_ARRAY(const uint8_t *, uint8, UINT8)
NVPAIR_CREATEV_ARRAY(const int16_t *, int16, INT16)
NVPAIR_CREATEV_ARRAY(const uint16_t *, uint16, UINT16)
NVPAIR_CREATEV_ARRAY(const int32_t *, int32, INT32)
NVPAIR_CREATEV_ARRAY(const uint32_t *, uint32, UINT32)
NVPAIR_CREATEV_ARRAY(const int64_t *, int64, INT64)
NVPAIR_CREATEV_ARRAY(const uint64_t *, uint64, UINT64)

nvpair_t *
nvpair_createv_string_array(const char * const *value, size_t nitems,
    const char *namefmt, va_list nameap)
{
	nvpair_t *nvp;
	unsigned int ii;
	size_t datasize, size;
	char **data, *ptr;

	PJDLOG_ASSERT((value == NULL && nitems == 0) ||
	    (value != NULL && nitems > 0));

	if (value == NULL) {
		data = NULL;
		datasize = 0;
	} else {
		datasize = 0;
		for (ii = 0; ii < nitems; ii++) {
			if (value[ii] != NULL)
				datasize += strlen(value[ii]) + 1;
		}
		data = malloc(sizeof(value[0]) * nitems + datasize);
		if (data == NULL)
			return (NULL);
		ptr = (char *)(data + nitems);
		for (ii = 0; ii < nitems; ii++) {
			if (value[ii] == NULL) {
				data[ii] = NULL;
			} else {
				size = strlen(value[ii]) + 1;
				memcpy(ptr, value[ii], size);
				data[ii] = ptr;
				ptr += size;
			}
		}
		PJDLOG_ASSERT((size_t)(ptr - (char *)data) ==
		    sizeof(value[0]) * nitems + datasize);
	}
	nvp = nvpair_allocv(NV_TYPE_STRING_ARRAY, (uint64_t)(uintptr_t)data,
	    sizeof(uint64_t) + sizeof(uint64_t) * nitems + datasize, nitems,
	    namefmt, nameap);
	if (nvp == NULL)
		free(data);
	return (nvp);
}

nvpair_t *
nvpair_createv_nvlist_array(const nvlist_t * const *value, size_t nitems,
    const char *namefmt, va_list nameap)
{
	unsigned int ii;
	nvlist_t **nvls;
	nvpair_t *nvp;

	PJDLOG_ASSERT((value == NULL && nitems == 0) ||
	    (value != NULL && nitems > 0));

	nvp = NULL;

	if (value == NULL) {
		nvls = NULL;
	} else {
		nvls = malloc(sizeof(value[0]) * nitems);
		if (nvls == NULL)
			return (NULL);
		for (ii = 0; ii < nitems; ii++) {
			if (value[ii] == NULL) {
				nvls[ii] = NULL;
			} else {
				nvls[ii] = nvlist_clone(value[ii]);
				if (nvls[ii] == NULL)
					goto fail;
			}
		}
	}
	nvp = nvpair_allocv(NV_TYPE_NVLIST_ARRAY, (uint64_t)(uintptr_t)nvls, 0,
	    nitems, namefmt, nameap);
	if (nvp == NULL && nvls != NULL) {
fail:
		for (; ii > 0; ii--)
			nvlist_destroy(nvls[ii - 1]);
		free(nvls);
	}
	return (nvp);
}

nvpair_t *
nvpair_createv_descriptor_array(const int *value, size_t nitems,
    const char *namefmt, va_list nameap)
{
	unsigned int ii;
	nvpair_t *nvp;
	int *fds;

	PJDLOG_ASSERT((value == NULL && nitems == 0) ||
	    (value != NULL && nitems > 0));

	nvp = NULL;

	if (value == NULL) {
		fds = NULL;
	} else {
		fds = malloc(sizeof(value[0]) * nitems);
		if (fds == NULL)
			return (NULL);
		for (ii = 0; ii < nitems; ii++) {
			if (value[ii] == -1) {
				fds[ii] = -1;
			} else {
				PJDLOG_ASSERT(fd_is_valid(value[ii]));

				fds[ii] = dup(value[ii]);
				if (fds[ii] == -1)
					goto fail;
			}
		}
	}
	nvp = nvpair_allocv(NV_TYPE_DESCRIPTOR_ARRAY, (uint64_t)(uintptr_t)fds,
	    sizeof(int64_t) * nitems, nitems, namefmt, nameap);
	if (nvp == NULL && fds != NULL) {
fail:
		for (; ii > 0; ii--) {
			if (fds[ii - 1] != -1)
				close(fds[ii - 1]);
		}
		free(fds);
	}
	return (nvp);
}

#define	NVPAIR_MOVE(vtype, type)					\
nvpair_t *								\
nvpair_move_##type(const char *name, vtype value)			\
{									\
									\
	return (nvpair_movef_##type(value, "%s", name));		\
}

NVPAIR_MOVE(char *, string)
NVPAIR_MOVE(nvlist_t *, nvlist)
NVPAIR_MOVE(int, descriptor)

#define	NVPAIR_MOVE_ARRAY(vtype, type)					\
nvpair_t *								\
nvpair_move_##type##_array(const char *name, vtype value, size_t nitems)\
{									\
									\
	return (nvpair_movef_##type##_array(value, nitems, "%s", name));\
}

NVPAIR_MOVE_ARRAY(bool *, bool)
NVPAIR_MOVE_ARRAY(int8_t *, int8)
NVPAIR_MOVE_ARRAY(uint8_t *, uint8)
NVPAIR_MOVE_ARRAY(int16_t *, int16)
NVPAIR_MOVE_ARRAY(uint16_t *, uint16)
NVPAIR_MOVE_ARRAY(int32_t *, int32)
NVPAIR_MOVE_ARRAY(uint32_t *, uint32)
NVPAIR_MOVE_ARRAY(int64_t *, int64)
NVPAIR_MOVE_ARRAY(uint64_t *, uint64)
NVPAIR_MOVE_ARRAY(char **, string)
NVPAIR_MOVE_ARRAY(nvlist_t **, nvlist)
NVPAIR_MOVE_ARRAY(int *, descriptor)

#define	NVPAIR_MOVEF(vtype, type)					\
nvpair_t *								\
nvpair_movef_##type(vtype value, const char *namefmt, ...)		\
{									\
	va_list nameap;							\
	nvpair_t *nvp;							\
									\
	va_start(nameap, namefmt);					\
	nvp = nvpair_movev_##type(value, namefmt, nameap);		\
	va_end(nameap);							\
									\
	return (nvp);							\
}

NVPAIR_MOVEF(char *, string)
NVPAIR_MOVEF(nvlist_t *, nvlist)
NVPAIR_MOVEF(int, descriptor)

#define	NVPAIR_MOVEF_ARRAY(vtype, type)					\
nvpair_t *								\
nvpair_movef_##type##_array(vtype value, size_t nitems,			\
    const char *namefmt, ...)						\
{									\
	va_list nameap;							\
	nvpair_t *nvp;							\
									\
	va_start(nameap, namefmt);					\
	nvp = nvpair_movev_##type##_array(value, nitems, namefmt,	\
	    nameap);							\
	va_end(nameap);							\
									\
	return (nvp);							\
}

NVPAIR_MOVEF_ARRAY(bool *, bool)
NVPAIR_MOVEF_ARRAY(int8_t *, int8)
NVPAIR_MOVEF_ARRAY(uint8_t *, uint8)
NVPAIR_MOVEF_ARRAY(int16_t *, int16)
NVPAIR_MOVEF_ARRAY(uint16_t *, uint16)
NVPAIR_MOVEF_ARRAY(int32_t *, int32)
NVPAIR_MOVEF_ARRAY(uint32_t *, uint32)
NVPAIR_MOVEF_ARRAY(int64_t *, int64)
NVPAIR_MOVEF_ARRAY(uint64_t *, uint64)
NVPAIR_MOVEF_ARRAY(char **, string)
NVPAIR_MOVEF_ARRAY(nvlist_t **, nvlist)
NVPAIR_MOVEF_ARRAY(int *, descriptor)

nvpair_t *
nvpair_movev_string(char *value, const char *namefmt, va_list nameap)
{
	nvpair_t *nvp;
	size_t size;

	size = value == NULL ? 0 : strlen(value) + 1;
	nvp = nvpair_allocv(NV_TYPE_STRING, (uint64_t)(uintptr_t)value, size, 0,
	    namefmt, nameap);
	if (nvp == NULL)
		free(value);
	return (nvp);
}

nvpair_t *
nvpair_movev_nvlist(nvlist_t *value, const char *namefmt, va_list nameap)
{
	nvpair_t *nvp;

	nvp = nvpair_allocv(NV_TYPE_NVLIST, (uint64_t)(uintptr_t)value, 0, 0,
	    namefmt, nameap);
	if (nvp == NULL)
		nvlist_destroy(value);
	return (nvp);
}

nvpair_t *
nvpair_movev_descriptor(int value, const char *namefmt, va_list nameap)
{

	PJDLOG_ASSERT(value == -1 || fd_is_valid(value));

	return (nvpair_allocv(NV_TYPE_DESCRIPTOR, (uint64_t)value,
	    sizeof(int64_t), 0, namefmt, nameap));
}

nvpair_t *
nvpair_movev_bool_array(bool *value, size_t nitems, const char *namefmt,
    va_list nameap)
{

	PJDLOG_ASSERT((value == NULL && nitems == 0) ||
	    (value != NULL && nitems > 0));

	return (nvpair_allocv(NV_TYPE_BOOL_ARRAY, (uint64_t)(uintptr_t)value,
	    sizeof(uint8_t) * nitems, nitems, namefmt, nameap));
}

#define	NVPAIR_MOVEV_ARRAY(type, TYPE)					\
nvpair_t *								\
nvpair_movev_##type##_array(type##_t *value, size_t nitems,		\
    const char *namefmt, va_list nameap)				\
{									\
									\
	PJDLOG_ASSERT((value == NULL && nitems == 0) ||			\
	    (value != NULL && nitems > 0));				\
									\
	return (nvpair_allocv(NV_TYPE_##TYPE##_ARRAY,			\
	    (uint64_t)(uintptr_t)value, sizeof(value[0]) * nitems,	\
	    nitems, namefmt, nameap));					\
}

NVPAIR_MOVEV_ARRAY(int8, INT8)
NVPAIR_MOVEV_ARRAY(uint8, UINT8)
NVPAIR_MOVEV_ARRAY(int16, INT16)
NVPAIR_MOVEV_ARRAY(uint16, UINT16)
NVPAIR_MOVEV_ARRAY(int32, INT32)
NVPAIR_MOVEV_ARRAY(uint32, UINT32)
NVPAIR_MOVEV_ARRAY(int64, INT64)
NVPAIR_MOVEV_ARRAY(uint64, UINT64)

nvpair_t *
nvpair_movev_string_array(char **value, size_t nitems, const char *namefmt,
    va_list nameap)
{
	size_t datasize, size;
	unsigned int ii;
	char *str;

	PJDLOG_ASSERT((value == NULL && nitems == 0) ||
	    (value != NULL && nitems > 0));

	/*
	 * Calculate datasize (sum of all strings len including
	 * terminating '\0').
	 * While doing so check if value is single buffer.
	 */
	datasize = 0;
	if (value != NULL) {
		str = (char *)(value + nitems);
		for (ii = 0; ii < nitems; ii++) {
			PJDLOG_ASSERT(value[ii] == NULL || value[ii] == str);

			if (value[ii] != NULL) {
				size = strlen(value[ii]) + 1;
				datasize += size;
				str += size;
			}
		}
	}

	return (nvpair_allocv(NV_TYPE_STRING_ARRAY, (uint64_t)(uintptr_t)value,
	    sizeof(uint64_t) + sizeof(uint64_t) * nitems + datasize, nitems,
	    namefmt, nameap));
}

nvpair_t *
nvpair_movev_nvlist_array(nvlist_t **value, size_t nitems, const char *namefmt,
    va_list nameap)
{

	PJDLOG_ASSERT((value == NULL && nitems == 0) ||
	    (value != NULL && nitems > 0));

	return (nvpair_allocv(NV_TYPE_NVLIST_ARRAY, (uint64_t)(uintptr_t)value,
	    0, nitems, namefmt, nameap));
}

nvpair_t *
nvpair_movev_descriptor_array(int *value, size_t nitems, const char *namefmt,
    va_list nameap)
{
	unsigned int ii;

	PJDLOG_ASSERT((value == NULL && nitems == 0) ||
	    (value != NULL && nitems > 0));

	if (value != NULL) {
		for (ii = 0; ii < nitems; ii++) {
			PJDLOG_ASSERT(value[ii] == -1 ||
			    fd_is_valid(value[ii]));
		}
	}
	return (nvpair_allocv(NV_TYPE_DESCRIPTOR_ARRAY,
	    (uint64_t)(uintptr_t)value, sizeof(int64_t) * nitems, nitems,
	    namefmt, nameap));
}

bool
nvpair_get_bool(const nvpair_t *nvp)
{

	NVPAIR_ASSERT(nvp);

	return (nvp->nvp_data == 1);
}

#define	NVPAIR_GET(type, TYPE)						\
type##_t								\
nvpair_get_##type(const nvpair_t *nvp)					\
{									\
									\
	NVPAIR_ASSERT(nvp);						\
	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_##TYPE);			\
									\
	return ((type##_t)nvp->nvp_data);				\
}

NVPAIR_GET(int8, INT8)
NVPAIR_GET(uint8, UINT8)
NVPAIR_GET(int16, INT16)
NVPAIR_GET(uint16, UINT16)
NVPAIR_GET(int32, INT32)
NVPAIR_GET(uint32, UINT32)
NVPAIR_GET(int64, INT64)
NVPAIR_GET(uint64, UINT64)

const char *
nvpair_get_string(const nvpair_t *nvp)
{

	NVPAIR_ASSERT(nvp);
	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_STRING);

	return ((const char *)(intptr_t)nvp->nvp_data);
}

const nvlist_t *
nvpair_get_nvlist(const nvpair_t *nvp)
{

	NVPAIR_ASSERT(nvp);
	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_NVLIST);

	return ((const nvlist_t *)(intptr_t)nvp->nvp_data);
}

int
nvpair_get_descriptor(const nvpair_t *nvp)
{

	NVPAIR_ASSERT(nvp);
	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_DESCRIPTOR);

	return ((int)nvp->nvp_data);
}

const bool *
nvpair_get_bool_array(const nvpair_t *nvp, size_t *nitemsp)
{

	NVPAIR_ASSERT(nvp);
	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_BOOL_ARRAY);

	if (nitemsp != NULL)
		*nitemsp = nvp->nvp_nitems;
	return ((const bool *)(intptr_t)nvp->nvp_data);
}

#define	NVPAIR_GET_ARRAY(type, TYPE)					\
const type##_t *							\
nvpair_get_##type##_array(const nvpair_t *nvp, size_t *nitemsp)		\
{									\
									\
	NVPAIR_ASSERT(nvp);						\
	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_##TYPE##_ARRAY);		\
									\
	if (nitemsp != NULL)						\
		*nitemsp = nvp->nvp_nitems;				\
	return ((const type##_t *)(intptr_t)nvp->nvp_data);		\
}

NVPAIR_GET_ARRAY(int8, INT8)
NVPAIR_GET_ARRAY(uint8, UINT8)
NVPAIR_GET_ARRAY(int16, INT16)
NVPAIR_GET_ARRAY(uint16, UINT16)
NVPAIR_GET_ARRAY(int32, INT32)
NVPAIR_GET_ARRAY(uint32, UINT32)
NVPAIR_GET_ARRAY(int64, INT64)
NVPAIR_GET_ARRAY(uint64, UINT64)

const char * const *
nvpair_get_string_array(const nvpair_t *nvp, size_t *nitemsp)
{

	NVPAIR_ASSERT(nvp);
	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_STRING_ARRAY);

	if (nitemsp != NULL)
		*nitemsp = nvp->nvp_nitems;
	return ((const char * const *)(intptr_t)nvp->nvp_data);
}

nvlist_t * const *
nvpair_get_nvlist_array(const nvpair_t *nvp, size_t *nitemsp)
{

	NVPAIR_ASSERT(nvp);
	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_NVLIST_ARRAY);

	if (nitemsp != NULL)
		*nitemsp = nvp->nvp_nitems;
	return ((nvlist_t * const *)(intptr_t)nvp->nvp_data);
}

const int *
nvpair_get_descriptor_array(const nvpair_t *nvp, size_t *nitemsp)
{

	NVPAIR_ASSERT(nvp);
	PJDLOG_ASSERT(nvp->nvp_type == NV_TYPE_DESCRIPTOR_ARRAY);

	if (nitemsp != NULL)
		*nitemsp = nvp->nvp_nitems;
	return ((const int *)(intptr_t)nvp->nvp_data);
}

void
nvpair_free(nvpair_t *nvp)
{
	unsigned int ii;
	nvlist_t **nvls;
	int *fds;

	NVPAIR_ASSERT(nvp);
	PJDLOG_ASSERT(nvp->nvp_list == NULL);

	nvp->nvp_magic = 0;
	switch (nvp->nvp_type) {
	case NV_TYPE_DESCRIPTOR:
		if ((int)nvp->nvp_data != -1)
			close((int)nvp->nvp_data);
		break;
	case NV_TYPE_NVLIST:
		nvlist_destroy((nvlist_t *)(intptr_t)nvp->nvp_data);
		break;
	case NV_TYPE_STRING:
		if (nvp->nvp_data != 0)
			free((char *)(intptr_t)nvp->nvp_data);
		break;
	case NV_TYPE_DESCRIPTOR_ARRAY:
		fds = (int *)(intptr_t)nvp->nvp_data;
		for (ii = 0; ii < nvp->nvp_nitems; ii++) {
			if (fds[ii] != -1)
				close(fds[ii]);
		}
		free(fds);
		break;
	case NV_TYPE_NVLIST_ARRAY:
		nvls = (nvlist_t **)(intptr_t)nvp->nvp_data;
		for (ii = 0; ii < nvp->nvp_nitems; ii++)
			nvlist_destroy(nvls[ii]);
		free(nvls);
		break;
	case NV_TYPE_BOOL_ARRAY:
	case NV_TYPE_INT8_ARRAY:
	case NV_TYPE_UINT8_ARRAY:
	case NV_TYPE_INT16_ARRAY:
	case NV_TYPE_UINT16_ARRAY:
	case NV_TYPE_INT32_ARRAY:
	case NV_TYPE_UINT32_ARRAY:
	case NV_TYPE_INT64_ARRAY:
	case NV_TYPE_UINT64_ARRAY:
	case NV_TYPE_STRING_ARRAY:
		if (nvp->nvp_data != 0)
			free((void *)(intptr_t)nvp->nvp_data);
		break;
	}
	free(nvp);
}

void
nvpair_free_structure(nvpair_t *nvp)
{

	NVPAIR_ASSERT(nvp);
	PJDLOG_ASSERT(nvp->nvp_list == NULL);

	nvp->nvp_magic = 0;
	free(nvp);
}

const char *
nvpair_type_string(const nvpair_t *nvp)
{

	NVPAIR_ASSERT(nvp);

	switch (nvp->nvp_type) {
	case NV_TYPE_NULL:
		return ("NULL");
	case NV_TYPE_BOOL:
		return ("BOOL");
	case NV_TYPE_INT8:
		return ("INT8");
	case NV_TYPE_UINT8:
		return ("UINT8");
	case NV_TYPE_INT16:
		return ("INT16");
	case NV_TYPE_UINT16:
		return ("UINT16");
	case NV_TYPE_INT32:
		return ("INT32");
	case NV_TYPE_UINT32:
		return ("UINT32");
	case NV_TYPE_INT64:
		return ("INT64");
	case NV_TYPE_UINT64:
		return ("UINT64");
	case NV_TYPE_STRING:
		return ("STRING");
	case NV_TYPE_NVLIST:
		return ("NVLIST");
	case NV_TYPE_DESCRIPTOR:
		return ("DESCRIPTOR");
	case NV_TYPE_BOOL_ARRAY:
		return ("BOOL_ARRAY");
	case NV_TYPE_INT8_ARRAY:
		return ("INT8_ARRAY");
	case NV_TYPE_UINT8_ARRAY:
		return ("UINT8_ARRAY");
	case NV_TYPE_INT16_ARRAY:
		return ("INT16_ARRAY");
	case NV_TYPE_UINT16_ARRAY:
		return ("UINT16_ARRAY");
	case NV_TYPE_INT32_ARRAY:
		return ("INT32_ARRAY");
	case NV_TYPE_UINT32_ARRAY:
		return ("UINT32_ARRAY");
	case NV_TYPE_INT64_ARRAY:
		return ("INT64_ARRAY");
	case NV_TYPE_UINT64_ARRAY:
		return ("UINT64_ARRAY");
	case NV_TYPE_STRING_ARRAY:
		return ("STRING_ARRAY");
	case NV_TYPE_NVLIST_ARRAY:
		return ("NVLIST_ARRAY");
	case NV_TYPE_DESCRIPTOR_ARRAY:
		return ("DESCRIPTOR_ARRAY");
	default:
		return ("<UNKNOWN>");
	}
}
